# yellowbrick.classifier.threshold
# DiscriminationThreshold visualizer for probabilistic classifiers.
#
# Author:  Nathan Danielsen
# Author:  Benjamin Bengfort
# Created: Wed April 26 20:17:29 2017 -0700
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: threshold.py [] nathan.danielsen@gmail.com $

"""
DiscriminationThreshold visualizer for probabilistic classifiers.
"""

##########################################################################
## Imports
##########################################################################

import bisect
import numpy as np

from scipy.stats import mstats
from collections import defaultdict

from sklearn.base import clone
from sklearn.utils import indexable
from sklearn.model_selection import ShuffleSplit
from sklearn.metrics import precision_recall_curve
from sklearn.utils.multiclass import type_of_target

try:
    # See #1137: this allows compatibility for scikit-learn >= 0.24
    from sklearn.utils import safe_indexing as _safe_indexing
except ImportError:
    from sklearn.utils import _safe_indexing

from yellowbrick.base import ModelVisualizer
from yellowbrick.style.colors import resolve_colors
from yellowbrick.utils import is_classifier, is_monotonic, is_probabilistic
from yellowbrick.exceptions import YellowbrickTypeError, YellowbrickValueError


# Quantiles for lower bound, curve, and upper bound
QUANTILES_MEDIAN_80 = np.array([0.1, 0.5, 0.9])

# List of threshold metrics
METRICS = ["precision", "recall", "fscore", "queue_rate"]


##########################################################################
# Discrimination Thresholds Visualization
##########################################################################


class DiscriminationThreshold(ModelVisualizer):
    """
    Visualizes how precision, recall, f1 score, and queue rate change as the
    discrimination threshold increases. For probabilistic, binary classifiers,
    the discrimination threshold is the probability at which you choose the
    positive class over the negative. Generally this is set to 50%, but
    adjusting the discrimination threshold will adjust sensitivity to false
    positives which is described by the inverse relationship of precision and
    recall with respect to the threshold.

    The visualizer also accounts for variability in the model by running
    multiple trials with different train and test splits of the data. The
    variability is visualized using a band such that the curve is drawn as the
    median score of each trial and the band is from the 10th to 90th
    percentile.

    The visualizer is intended to help users determine an appropriate
    threshold for decision making (e.g. at what threshold do we have a human
    review the data), given a tolerance for precision and recall or limiting
    the number of records to check (the queue rate).

    Parameters
    ----------
    estimator : estimator
        A scikit-learn estimator that should be a classifier. If the model is
        not a classifier, an exception is raised. If the internal model is not
        fitted, it is fit when the visualizer is fitted, unless otherwise specified
        by ``is_fitted``.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If not specified the current axes will be
        used (or generated if required).

    n_trials : integer, default: 50
        Number of times to shuffle and split the dataset to account for noise
        in the threshold metrics curves. Note if cv provides > 1 splits,
        the number of trials will be n_trials * cv.get_n_splits()

    cv : float or cross-validation generator, default: 0.1
        Determines the splitting strategy for each trial. Possible inputs are:

        - float, to specify the percent of the test split
        - object to be used as cross-validation generator

        This attribute is meant to give flexibility with stratified splitting
        but if a splitter is provided, it should only return one split and
        have shuffle set to True.

    fbeta : float, 1.0 by default
        The strength of recall versus precision in the F-score.

    argmax : str or None, default: 'fscore'
        Annotate the threshold maximized by the supplied metric (see exclude
        for the possible metrics to use). If None or passed to exclude,
        will not annotate the graph.

    exclude : str or list, optional
        Specify metrics to omit from the graph, can include:

        - ``"precision"``
        - ``"recall"``
        - ``"queue_rate"``
        - ``"fscore"``

        Excluded metrics will not be displayed in the graph, nor will they
        be available in ``thresholds_``; however, they will be computed on fit.

    quantiles : sequence, default: np.array([0.1, 0.5, 0.9])
        Specify the quantiles to view model variability across a number of
        trials. Must be monotonic and have three elements such that the first
        element is the lower bound, the second is the drawn curve, and the
        third is the upper bound. By default the curve is drawn at the median,
        and the bounds from the 10th percentile to the 90th percentile.

    random_state : int, optional
        Used to seed the random state for shuffling the data while composing
        different train and test splits. If supplied, the random state is
        incremented in a deterministic fashion for each split.

        Note that if a splitter is provided, it's random state will also be
        updated with this random state, even if it was previously set.

    is_fitted : bool or str, default="auto"
        Specify if the wrapped estimator is already fitted. If False, the estimator
        will be fit when the visualizer is fit, otherwise, the estimator will not be
        modified. If "auto" (default), a helper method will check if the estimator
        is fitted before fitting it again.

    force_model : bool, default: False
        Do not check to ensure that the underlying estimator is a classifier. This
        will prevent an exception when the visualizer is initialized but may result
        in unexpected or unintended behavior.

    kwargs : dict
        Keyword arguments passed to the visualizer base classes.

    Attributes
    ----------
    thresholds_ : array
        The uniform thresholds identified by each of the trial runs.

    cv_scores_ : dict of arrays of ``len(thresholds_)``
        The values for all included metrics including the upper and lower
        bounds of the metrics defined by quantiles.

    Notes
    -----
    The term "discrimination threshold" is rare in the literature. Here, we
    use it to mean the probability at which the positive class is selected
    over the negative class in binary classification.

    Classification models must implement either a ``decision_function`` or
    ``predict_proba`` method in order to be used with this class. A
    ``YellowbrickTypeError`` is raised otherwise.

    .. caution:: This method only works for binary, probabilistic classifiers.

    .. seealso::
        For a thorough explanation of discrimination thresholds, see:
        `Visualizing Machine Learning Thresholds to Make Better Business
        Decisions
        <http://blog.insightdatalabs.com/visualizing-classifier-thresholds/>`_
        by Insight Data.
    """

    def __init__(
        self,
        estimator,
        ax=None,
        n_trials=50,
        cv=0.1,
        fbeta=1.0,
        argmax="fscore",
        exclude=None,
        quantiles=QUANTILES_MEDIAN_80,
        random_state=None,
        is_fitted="auto",
        force_model=False,
        **kwargs
    ):

        # Perform some quick type checking to help users avoid error.
        if not force_model and (
            not is_classifier(estimator) or not is_probabilistic(estimator)
        ):
            raise YellowbrickTypeError(
                "{} requires a probabilistic binary classifier".format(
                    self.__class__.__name__
                )
            )

        # Check the various inputs
        self._check_quantiles(quantiles)
        self._check_cv(cv)
        self._check_exclude(exclude)
        self._check_argmax(argmax, exclude)

        # Initialize the ModelVisualizer
        super(DiscriminationThreshold, self).__init__(
            estimator, ax=ax, is_fitted=is_fitted, **kwargs
        )

        # Set params
        self.n_trials = n_trials
        self.cv = cv
        self.fbeta = fbeta
        self.argmax = argmax
        self.exclude = exclude
        self.quantiles = quantiles
        self.random_state = random_state

    def fit(self, X, y, **kwargs):
        """
        Fit is the entry point for the visualizer. Given instances described
        by X and binary classes described in the target y, fit performs n
        trials by shuffling and splitting the dataset then computing the
        precision, recall, f1, and queue rate scores for each trial. The
        scores are aggregated by the quantiles expressed then drawn.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features

        y : ndarray or Series of length n
            An array or series of target or class values. The target y must
            be a binary classification target.

        kwargs: dict
            keyword arguments passed to Scikit-Learn API.

        Returns
        -------
        self : instance
            Returns the instance of the visualizer

        raises: YellowbrickValueError
            If the target y is not a binary classification target.
        """
        # Check target before metrics raise crazy exceptions
        # TODO: Switch to using target type from utils.target
        if type_of_target(y) != "binary":
            raise YellowbrickValueError("multiclass format is not supported")

        # Make arrays indexable for cross validation
        X, y = indexable(X, y)

        # TODO: parallelize trials with joblib (using sklearn utility)
        # NOTE: parallelization with matplotlib is tricky at best!
        trials = [
            metric
            for idx in range(self.n_trials)
            for metric in self._split_fit_score_trial(X, y, idx)
        ]

        # Compute maximum number of uniform thresholds across all trials
        n_thresholds = np.array([len(t["thresholds"]) for t in trials]).min()
        self.thresholds_ = np.linspace(0.0, 1.0, num=n_thresholds)

        # Filter metrics and collect values for uniform thresholds
        metrics = frozenset(METRICS) - self._check_exclude(self.exclude)
        uniform_metrics = defaultdict(list)

        for trial in trials:
            rows = defaultdict(list)
            for t in self.thresholds_:
                idx = bisect.bisect_left(trial["thresholds"], t)
                for metric in metrics:
                    rows[metric].append(trial[metric][idx])

            for metric, row in rows.items():
                uniform_metrics[metric].append(row)

        # Convert metrics to metric arrays
        uniform_metrics = {
            metric: np.array(values) for metric, values in uniform_metrics.items()
        }

        # Perform aggregation and store cv_scores_
        quantiles = self._check_quantiles(self.quantiles)
        self.cv_scores_ = {}

        for metric, values in uniform_metrics.items():
            # Compute the lower, median, and upper plots
            lower, median, upper = mstats.mquantiles(values, prob=quantiles, axis=0)

            # Store the aggregates in cv scores
            self.cv_scores_[metric] = median
            self.cv_scores_["{}_lower".format(metric)] = lower
            self.cv_scores_["{}_upper".format(metric)] = upper

        # TODO: fit the underlying estimator with the best decision threshold
        # Call super to ensure the underlying estimator is correctly fitted
        super(DiscriminationThreshold, self).fit(X, y)

        # Draw and always return self
        self.draw()
        return self

    def _split_fit_score_trial(self, X, y, idx=0):
        """
        Splits the dataset, fits a clone of the estimator, then scores it
        according to the required metrics.

        The index of the split is added to the random_state if the
        random_state is not None; this ensures that every split is shuffled
        differently but in a deterministic fashion for testing purposes.
        """
        random_state = self.random_state
        if random_state is not None:
            random_state += idx

        splitter = self._check_cv(self.cv, random_state)

        for train_index, test_index in splitter.split(X, y):
            # Safe indexing handles multiple types of inputs including
            # DataFrames and structured arrays - required for generic splits.
            X_train = _safe_indexing(X, train_index)
            y_train = _safe_indexing(y, train_index)
            X_test = _safe_indexing(X, test_index)
            y_test = _safe_indexing(y, test_index)

            model = clone(self.estimator)
            model.fit(X_train, y_train)

            if hasattr(model, "predict_proba"):
                # Get the probabilities for the positive class
                y_scores = model.predict_proba(X_test)[:, 1]
            else:
                # Use the decision function to get the scores
                y_scores = model.decision_function(X_test)

            # Compute the curve metrics and thresholds
            curve_metrics = precision_recall_curve(y_test, y_scores)
            precision, recall, thresholds = curve_metrics

            # Compute the F1 score from precision and recall
            # Don't need to warn for F, precision/recall would have warned
            with np.errstate(divide="ignore", invalid="ignore"):
                beta = self.fbeta ** 2
                f_score = (1 + beta) * precision * recall / (beta * precision + recall)

            # Ensure thresholds ends at 1
            thresholds = np.append(thresholds, 1)

            # Compute the queue rate
            queue_rate = np.array(
                [(y_scores >= threshold).mean() for threshold in thresholds]
            )

            yield {
                "thresholds": thresholds,
                "precision": precision,
                "recall": recall,
                "fscore": f_score,
                "queue_rate": queue_rate,
            }

    def draw(self):
        """
        Draws the cv scores as a line chart on the current axes.
        """
        # Set the colors from the supplied values or reasonable defaults
        color_values = resolve_colors(n_colors=4, colors=self.color)

        # Get the metric used to annotate the graph with its maximizing value
        argmax = self._check_argmax(self.argmax, self.exclude)

        for idx, metric in enumerate(METRICS):
            # Skip any excluded labels
            if metric not in self.cv_scores_:
                continue

            # Get the color ensuring every metric has a static color
            color = color_values[idx]

            # Make the label pretty
            if metric == "fscore":
                if self.fbeta == 1.0:
                    label = "$f_1$"
                else:
                    label = "$f_{{\beta={:0.1f}}}".format(self.fbeta)
            else:
                label = metric.replace("_", " ")

            # Draw the metric values
            self.ax.plot(
                self.thresholds_, self.cv_scores_[metric], color=color, label=label
            )

            # Draw the upper and lower bounds
            lower = self.cv_scores_["{}_lower".format(metric)]
            upper = self.cv_scores_["{}_upper".format(metric)]

            self.ax.fill_between(
                self.thresholds_, upper, lower, alpha=0.35, linewidth=0, color=color
            )

            # Annotate the graph with the maximizing value
            if argmax and argmax == metric:
                argmax = self.cv_scores_[metric].argmax()
                threshold = self.thresholds_[argmax]
                self.ax.axvline(
                    threshold,
                    ls="--",
                    c="k",
                    lw=1,
                    label="$t_{}={:0.2f}$".format(metric[0], threshold),
                )

        return self.ax

    def finalize(self, **kwargs):
        """
        Sets a title and axis labels on the visualizer and ensures that the
        axis limits are scaled to valid threshold values.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        Notes
        -----
        Generally this method is called from show and not directly by the user.
        """
        super(DiscriminationThreshold, self).finalize(**kwargs)

        # Set the title of the threshold visualiztion
        self.set_title("Threshold Plot for {}".format(self.name))

        self.ax.legend(frameon=True, loc="best")
        self.ax.set_xlabel("discrimination threshold")
        self.ax.set_ylabel("score")
        self.ax.set_xlim(0.0, 1.0)
        self.ax.set_ylim(0.0, 1.0)

    def _check_quantiles(self, val):
        """
        Validate the quantiles passed in. Returns the np array if valid.
        """
        if len(val) != 3 or not is_monotonic(val) or not np.all(val < 1):
            raise YellowbrickValueError(
                "quantiles must be a sequence of three "
                "monotonically increasing values less than 1"
            )
        return np.asarray(val)

    def _check_cv(self, val, random_state=None):
        """
        Validate the cv method passed in. Returns the split strategy if no
        validation exception is raised.
        """
        # Use default splitter in this case
        if val is None:
            val = 0.1

        if isinstance(val, float) and val <= 1.0:
            return ShuffleSplit(n_splits=1, test_size=val, random_state=random_state)

        if hasattr(val, "split") and hasattr(val, "get_n_splits"):
            if random_state is not None and hasattr(val, "random_state"):
                val.random_state = random_state
            return val

        raise YellowbrickValueError("'{}' is not a valid cv splitter".format(type(val)))

    def _check_exclude(self, val):
        """
        Validate the excluded metrics. Returns the set of excluded params.
        """
        if val is None:
            exclude = frozenset()
        elif isinstance(val, str):
            exclude = frozenset([val.lower()])
        else:
            exclude = frozenset(map(lambda s: s.lower(), val))

        if len(exclude - frozenset(METRICS)) > 0:
            raise YellowbrickValueError(
                "'{}' is not a valid metric to exclude".format(repr(val))
            )

        return exclude

    def _check_argmax(self, val, exclude=None):
        """
        Validate the argmax metric. Returns the metric used to annotate the graph.
        """
        if val is None:
            return None

        argmax = val.lower()

        if argmax not in METRICS:
            raise YellowbrickValueError(
                "'{}' is not a valid metric to use".format(repr(val))
            )

        exclude = self._check_exclude(exclude)
        if argmax in exclude:
            argmax = None

        return argmax


##########################################################################
# Quick Methods
##########################################################################

def discrimination_threshold(
    estimator,
    X,
    y,
    ax=None,
    n_trials=50,
    cv=0.1,
    fbeta=1.0,
    argmax="fscore",
    exclude=None,
    quantiles=QUANTILES_MEDIAN_80,
    random_state=None,
    is_fitted="auto",
    force_model=False,
    show=True,
    **kwargs
):
    """Discrimination Threshold

    Visualizes how precision, recall, f1 score, and queue rate change as the
    discrimination threshold increases. For probabilistic, binary classifiers,
    the discrimination threshold is the probability at which you choose the
    positive class over the negative. Generally this is set to 50%, but
    adjusting the discrimination threshold will adjust sensitivity to false
    positives which is described by the inverse relationship of precision and
    recall with respect to the threshold.

    .. seealso:: See DiscriminationThreshold for more details.

    Parameters
    ----------
    estimator : estimator
        A scikit-learn estimator that should be a classifier. If the model is
        not a classifier, an exception is raised. If the internal model is not
        fitted, it is fit when the visualizer is fitted, unless otherwise specified
        by ``is_fitted``.

    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features

    y : ndarray or Series of length n
        An array or series of target or class values. The target y must
        be a binary classification target.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If not specified the current axes will be
        used (or generated if required).

    n_trials : integer, default: 50
        Number of times to shuffle and split the dataset to account for noise
        in the threshold metrics curves. Note if cv provides > 1 splits,
        the number of trials will be n_trials * cv.get_n_splits()

    cv : float or cross-validation generator, default: 0.1
        Determines the splitting strategy for each trial. Possible inputs are:

        - float, to specify the percent of the test split
        - object to be used as cross-validation generator

        This attribute is meant to give flexibility with stratified splitting
        but if a splitter is provided, it should only return one split and
        have shuffle set to True.

    fbeta : float, 1.0 by default
        The strength of recall versus precision in the F-score.

    argmax : str or None, default: 'fscore'
        Annotate the threshold maximized by the supplied metric (see exclude
        for the possible metrics to use). If None or passed to exclude,
        will not annotate the graph.

    exclude : str or list, optional
        Specify metrics to omit from the graph, can include:

        - ``"precision"``
        - ``"recall"``
        - ``"queue_rate"``
        - ``"fscore"``

        Excluded metrics will not be displayed in the graph, nor will they
        be available in ``thresholds_``; however, they will be computed on fit.

    quantiles : sequence, default: np.array([0.1, 0.5, 0.9])
        Specify the quantiles to view model variability across a number of
        trials. Must be monotonic and have three elements such that the first
        element is the lower bound, the second is the drawn curve, and the
        third is the upper bound. By default the curve is drawn at the median,
        and the bounds from the 10th percentile to the 90th percentile.

    random_state : int, optional
        Used to seed the random state for shuffling the data while composing
        different train and test splits. If supplied, the random state is
        incremented in a deterministic fashion for each split.

        Note that if a splitter is provided, it's random state will also be
        updated with this random state, even if it was previously set.

    is_fitted : bool or str, default="auto"
        Specify if the wrapped estimator is already fitted. If False, the estimator
        will be fit when the visualizer is fit, otherwise, the estimator will not be
        modified. If "auto" (default), a helper method will check if the estimator
        is fitted before fitting it again.

    force_model : bool, default: False
        Do not check to ensure that the underlying estimator is a classifier. This
        will prevent an exception when the visualizer is initialized but may result
        in unexpected or unintended behavior.

    show : bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``

    kwargs : dict
        Keyword arguments passed to the visualizer base classes.

    Notes
    -----
    The term "discrimination threshold" is rare in the literature. Here, we
    use it to mean the probability at which the positive class is selected
    over the negative class in binary classification.

    Classification models must implement either a ``decision_function`` or
    ``predict_proba`` method in order to be used with this class. A
    ``YellowbrickTypeError`` is raised otherwise.

    .. seealso::
        For a thorough explanation of discrimination thresholds, see:
        `Visualizing Machine Learning Thresholds to Make Better Business
        Decisions
        <http://blog.insightdatalabs.com/visualizing-classifier-thresholds/>`_
        by Insight Data.

    Examples
    --------
    >>> from yellowbrick.classifier.threshold import discrimination_threshold
    >>> from sklearn.linear_model import LogisticRegression
    >>> from yellowbrick.datasets import load_occupancy
    >>> X, y = load_occupancy()
    >>> model = LogisticRegression(multi_class="auto", solver="liblinear")
    >>> discrimination_threshold(model, X, y)

    Returns
    -------
    viz : DiscriminationThreshold
        Returns the fitted and finalized visualizer object.
    """
    # Instantiate the visualizer
    visualizer = DiscriminationThreshold(
        estimator,
        ax=ax,
        n_trials=n_trials,
        cv=cv,
        fbeta=fbeta,
        argmax=argmax,
        exclude=exclude,
        quantiles=quantiles,
        random_state=random_state,
        is_fitted=is_fitted,
        force_model=force_model,
        **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(X, y)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer
    return visualizer



# yellowbrick.cluster.base
# Base class for cluster visualizers.
#
# Author:   Benjamin Bengfort
# Created:  Thu Mar 23 17:28:38 2017 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: base.py [241edca] benjamin@bengfort.com $

"""
Base class for cluster visualizers.
"""

##########################################################################
## Imports
##########################################################################

from yellowbrick.utils import isclusterer
from yellowbrick.base import ScoreVisualizer
from yellowbrick.exceptions import YellowbrickTypeError


## Packages for export
__all__ = ["ClusteringScoreVisualizer"]


##########################################################################
## Clustering Score Visualization Base Object
##########################################################################


class ClusteringScoreVisualizer(ScoreVisualizer):
    """
    Base class for all ScoreVisualizers that evaluate a clustering estimator.

    The primary functionality of this class is to perform a check to ensure
    that the wrapped estimator is a cluster estimator, otherwise a
    ``YellowbrickTypeError`` exception is raised.
    """

    def __init__(self, estimator, ax=None, fig=None, force_model=False, **kwargs):
        if not force_model and not isclusterer(estimator):
            raise YellowbrickTypeError(
                "The supplied model is not a clustering estimator; try a "
                "classifier or regression score visualizer instead!"
            )
        self.force_model = force_model
        super(ClusteringScoreVisualizer, self).__init__(
            estimator, ax=ax, fig=fig, **kwargs
        )



# yellowbrick.classifier.base
# API for classification visualizer hierarchy.
#
# Author:   Rebecca Bilbro
# Author:   Benjamin Bengfort
# Author:   Neal Humphrey
# Created:  Wed May 18 12:39:40 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: base.py [5388065] neal@nhumphrey.com $

"""
API for classification visualizer hierarchy.
"""

##########################################################################
## Imports
##########################################################################

import warnings
import numpy as np

from yellowbrick.utils import isclassifier
from yellowbrick.base import ScoreVisualizer
from yellowbrick.style.palettes import color_palette
from yellowbrick.exceptions import NotFitted, YellowbrickWarning
from yellowbrick.exceptions import YellowbrickTypeError, ModelError

from sklearn.preprocessing import LabelEncoder


##########################################################################
## Base Classification Visualizer
##########################################################################


class ClassificationScoreVisualizer(ScoreVisualizer):
    """Base class for classifier model selection.

    The ClassificationScoreVisualizer wraps a classifier to produce a
    visualization when the score method is called, usually to allow the user
    to effectively compare the performance between models.

    The base class provides helper functionality to ensure that classification
    visualizers are able to correctly identify and encode classes with human
    readable labels and to map colors to the classes if required.

    Parameters
    ----------
    estimator : estimator
        A scikit-learn estimator that should be a classifier. If the model is
        not a classifier, an exception is raised. If the internal model is not
        fitted, it is fit when the visualizer is fitted, unless otherwise specified
        by ``is_fitted``.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If not specified the current axes will be
        used (or generated if required).

    fig : matplotlib Figure, default: None
        The figure to plot the Visualizer on. If None is passed in the current
        plot will be used (or generated if required).

    classes : list of str, defult: None
        The class labels to use for the legend ordered by the index of the sorted
        classes discovered in the ``fit()`` method. Specifying classes in this
        manner is used to change the class names to a more specific format or
        to label encoded integer classes. Some visualizers may also use this
        field to filter the visualization for specific classes. For more advanced
        usage specify an encoder rather than class labels.

    encoder : dict or LabelEncoder, default: None
        A mapping of classes to human readable labels. Often there is a mismatch
        between desired class labels and those contained in the target variable
        passed to ``fit()`` or ``score()``. The encoder disambiguates this mismatch
        ensuring that classes are labeled correctly in the visualization.

    is_fitted : bool or str, default="auto"
        Specify if the wrapped estimator is already fitted. If False, the estimator
        will be fit when the visualizer is fit, otherwise, the estimator will not be
        modified. If "auto" (default), a helper method will check if the estimator
        is fitted before fitting it again.

    force_model : bool, default: False
        Do not check to ensure that the underlying estimator is a classifier. This
        will prevent an exception when the visualizer is initialized but may result
        in unexpected or unintended behavior.

    kwargs : dict
        Keyword arguments passed to the visualizer base classes.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        The class labels observed while fitting.

    class_count_ : ndarray of shape (n_classes,)
        Number of samples encountered for each class during fitting.

    score_ : float
        An evaluation metric of the classifier on test data produced when
        ``score()`` is called. This metric is between 0 and 1 -- higher scores are
        generally better. For classifiers, this score is usually accuracy, but
        ensure you check the underlying model for more details about the metric.
    """

    def __init__(
        self,
        estimator,
        ax=None,
        fig=None,
        classes=None,
        encoder=None,
        is_fitted="auto",
        force_model=False,
        **kwargs
    ):
        # A bit of type checking
        if not force_model and not isclassifier(estimator):
            raise YellowbrickTypeError(
                "This estimator is not a classifier; "
                "try a regression or clustering score visualizer instead!"
            )

        # Initialize the super method.
        super(ClassificationScoreVisualizer, self).__init__(
            estimator, ax=ax, fig=fig, is_fitted=is_fitted, **kwargs
        )

        self.classes = classes
        self.encoder = encoder
        self.force_model = force_model

    @property
    def class_colors_(self):
        """
        Returns ``_colors`` if it exists, otherwise computes a categorical color
        per class based on the matplotlib color cycle. If the visualizer is not
        fitted, raises a NotFitted exception.

        If subclasses require users to choose colors or have specialized color
        handling, they should set ``_colors`` on init or during fit.

        Notes
        -----
        Because this is a property, this docstring is for developers only.
        """
        if not hasattr(self, "_colors"):
            if not hasattr(self, "classes_"):
                raise NotFitted("cannot determine colors before fit")

            # TODO: replace with resolve_colors
            self._colors = color_palette(None, len(self.classes_))
        return self._colors

    def fit(self, X, y=None, **kwargs):
        """
        Fit the visualizer to the specified data.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features

        y : ndarray or Series of length n
            An array or series of target or class values

        Returns
        -------
        self : instance
            Returns the instance of the classification score visualizer
        """
        # Super fits the wrapped estimator
        super(ClassificationScoreVisualizer, self).fit(X, y, **kwargs)

        # Extract the classes and the class counts from the target
        self.classes_, self.class_counts_ = np.unique(y, return_counts=True)

        # Ensure the classes are aligned with the estimator
        # If they are not aligned, ignore class counts and issue a warning
        if hasattr(self.estimator, "classes_"):
            if not np.array_equal(self.classes_, self.estimator.classes_):
                self.classes_ = self.estimator.classes_
                self.class_counts_ = None

        # Decode classes to human readable labels specified by the user
        self.classes_ = self._decode_labels(self.classes_)

        # Always return self from fit
        return self

    def score(self, X, y):
        """
        The score function is the hook for visual interaction. Pass in test
        data and the visualizer will create predictions on the data and
        evaluate them with respect to the test values. The evaluation will
        then be passed to draw() and the result of the estimator score will
        be returned.

        Parameters
        ----------
        X : array-like
            X (also X_test) are the dependent variables of test set to predict

        y : array-like
            y (also y_test) is the independent actual variables to score against

        Returns
        -------
        score : float
            Returns the score of the underlying model, usually accuracy for
            classification models. Refer to the specific model for more details.
        """
        # If the estimator has been passed in fitted but the visualizer was not fit
        # then we can retrieve the classes from the estimator, unfortunately we cannot
        # retrieve the class counts so we simply set them to None and warn the user.
        # NOTE: cannot test if hasattr(self, "classes_") because it will be proxied.
        if not hasattr(self, "class_counts_"):
            if not hasattr(self.estimator, "classes_"):
                raise NotFitted(
                    (
                        "could not determine required property classes_; "
                        "the visualizer must either be fit or instantiated with a "
                        "fitted classifier before calling score()"
                    )
                )

            self.class_counts_ = None
            self.classes_ = self._decode_labels(self.estimator.classes_)
            warnings.warn(
                "could not determine class_counts_ from previously fitted classifier",
                YellowbrickWarning,
            )

        # This method implements ScoreVisualizer (do not call super).
        self.score_ = self.estimator.score(X, y)
        return self.score_

    def _decode_labels(self, y):
        """
        An internal helper function that uses either the classes or encoder
        properties to correctly decode y as user-readable string labels.

        If both classes and encoder are set, a warning is issued and encoder is
        used instead of classes. If neither encoder nor classes is set then the
        original array is returned unmodified.
        """
        if self.classes is not None and self.encoder is not None:
            warnings.warn(
                "both classes and encoder specified, using encoder", YellowbrickWarning
            )

        if self.encoder is not None:
            # Use the label encoder or other transformer
            if hasattr(self.encoder, "inverse_transform"):
                try:
                    return self.encoder.inverse_transform(y)
                except ValueError:
                    y_labels = np.unique(y)
                    raise ModelError(
                        "could not decode {} y values to {} labels".format(
                            y_labels, self._labels()
                        )
                    )

            # Otherwise, treat as a dictionary
            try:
                return np.asarray([self.encoder[yi] for yi in y])
            except KeyError as e:
                raise ModelError(
                    (
                        "cannot decode class {} to label, "
                        "key not specified by encoder"
                    ).format(e)
                )

        if self.classes is not None:
            # Determine indices to perform class mappings on
            yp = np.asarray(y)
            if yp.dtype.kind in {"i", "u"}:
                idx = yp
            else:
                # Use label encoder to get indices by sorted class names
                idx = LabelEncoder().fit_transform(yp)

            # Use index mapping for classes
            try:
                return np.asarray(self.classes)[idx]
            except IndexError:
                y_labels = np.unique(yp)
                raise ModelError(
                    "could not decode {} y values to {} labels".format(
                        y_labels, self._labels()
                    )
                )

        # could not decode y without encoder or classes, return it as it is, unmodified
        return y

    def _labels(self):
        """
        Returns the human specified labels in either the classes list or from the
        encoder. Returns None if no human labels have been specified, but issues a
        warning if a transformer has been passed that does not specify labels.
        """
        if self.classes is not None and self.encoder is not None:
            warnings.warn(
                "both classes and encoder specified, using encoder", YellowbrickWarning
            )

        if self.encoder is not None:
            # Use label encoder or other transformer
            if hasattr(self.encoder, "transform"):
                if hasattr(self.encoder, "classes_"):
                    return self.encoder.classes_

                # This is not a label encoder
                msg = "could not determine class labels from {}".format(
                    self.encoder.__class__.__name__
                )
                warnings.warn(msg, YellowbrickWarning)
                return None

            # Otherwise, treat as dictionary and ensure sorted by key
            keys = sorted(list(self.encoder.keys()))
            return np.asarray([self.encoder[key] for key in keys])

        if self.classes is not None:
            return np.asarray(self.classes)

        return None



# yellowbrick.utils.decorators
# Decorators and descriptors for annotating yellowbrick library functions.
#
# Author:   Benjamin Bengfort
# Created:  Thu May 18 15:13:33 2017 -0400
#
# Copyright (C) 2017 The sckit-yb developers
# For license information, see LICENSE.txt
#
# ID: decorators.py [79cd8cf] benjamin@bengfort.com $

"""
Decorators and descriptors for annotating yellowbrick library functions.
"""

##########################################################################
## Imports
##########################################################################

from functools import wraps


##########################################################################
## Decorators
##########################################################################


def memoized(fget):
    """
    Return a property attribute for new-style classes that only calls its
    getter on the first access. The result is stored and on subsequent
    accesses is returned, preventing the need to call the getter any more.

    Parameters
    ----------
    fget: function
        The getter method to memoize for subsequent access.

    See also
    --------
    python-memoized-property
        `python-memoized-property <https://github.com/estebistec/python-memoized-property>`_
    """
    attr_name = "_{0}".format(fget.__name__)

    @wraps(fget)
    def fget_memoized(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fget(self))
        return getattr(self, attr_name)

    return property(fget_memoized)


class docutil(object):
    """
    This decorator can be used to apply the doc string from another function
    to the decorated function. This is used for our single call wrapper
    functions who implement the visualizer API without forcing the user to
    jump through all the hoops. The docstring of both the visualizer and the
    single call wrapper should be identical, this decorator ensures that we
    only have to edit one doc string.

    Usage::

        @docutil(Visualizer.__init__)
        def visualize(*args, **kwargs):
            pass

    The basic usage is that you instantiate the decorator with the function
    whose docstring you want to copy, then apply that decorator to the the
    function whose docstring you would like modified.

    Note that this decorator performs no wrapping of the target function.
    """

    def __init__(self, func):
        """Create a decorator to document other functions with the specified
        function's doc string.

        Parameters
        ----------
        func : function
            The function whose doc string we should decorate with
        """
        self.doc = func.__doc__

    def __call__(self, func):
        """Modify the decorated function with the stored doc string.

        Parameters
        ----------
        func: function
            The function to apply the saved doc string to.
        """
        func.__doc__ = self.doc
        return func



# yellowbrick.cluster.icdm
# Implements Intercluster Distance Map visualizations.
#
# Author:  Benjamin Bengfort
# Created: Tue Aug 21 11:56:53 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: icdm.py [2f23976] benjamin@bengfort.com $

"""
Implements Intercluster Distance Map visualizations.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np
import matplotlib.pyplot as plt

from matplotlib.patches import Circle
from sklearn.manifold import MDS, TSNE

from yellowbrick.utils.timer import Timer
from yellowbrick.utils.decorators import memoized
from yellowbrick.exceptions import YellowbrickValueError
from yellowbrick.cluster.base import ClusteringScoreVisualizer
from yellowbrick.utils.helpers import prop_to_size, check_fitted

try:
    # Only available in Matplotlib >= 2.0.2
    from mpl_toolkits.axes_grid1 import inset_locator
except ImportError:
    inset_locator = None


## Packages for export
__all__ = [
    "InterclusterDistance",
    "intercluster_distance",
    "VALID_EMBEDDING",
    "VALID_SCORING",
    "ICDM",
]


# Valid strings to use for embedding names
VALID_EMBEDDING = {"mds", "tsne"}

# Valid strings to use for scoring names
VALID_SCORING = {"membership"}


##########################################################################
## InterclusterDistance Visualizer
##########################################################################


class InterclusterDistance(ClusteringScoreVisualizer):
    """
    Intercluster distance maps display an embedding of the cluster centers in
    2 dimensions with the distance to other centers preserved. E.g. the closer
    to centers are in the visualization, the closer they are in the original
    feature space. The clusters are sized according to a scoring metric. By
    default, they are sized by membership, e.g. the number of instances that
    belong to each center. This gives a sense of the relative importance of
    clusters. Note however, that because two clusters overlap in the 2D space,
    it does not imply that they overlap in the original feature space.

    Parameters
    ----------
    estimator : a Scikit-Learn clusterer
        Should be an instance of a centroidal clustering algorithm (or a
        hierarchical algorithm with a specified number of clusters). Also
        accepts some other models like LDA for text clustering.
        If it is not a clusterer, an exception is raised. If the estimator
        is not fitted, it is fit when the visualizer is fitted, unless
        otherwise specified by ``is_fitted``.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    min_size : int, default: 400
        The size, in points, of the smallest cluster drawn on the graph.
        Cluster sizes will be scaled between the min and max sizes.

    max_size : int, default: 25000
        The size, in points, of the largest cluster drawn on the graph.
        Cluster sizes will be scaled between the min and max sizes.

    embedding : default: 'mds'
        The algorithm used to embed the cluster centers in 2 dimensional space
        so that the distance between clusters is represented equivalently to
        their relationship in feature spaceself.
        Embedding algorithm options include:

        - **mds**: multidimensional scaling
        - **tsne**: stochastic neighbor embedding

    scoring : default: 'membership'
        The scoring method used to determine the size of the clusters drawn on
        the graph so that the relative importance of clusters can be viewed.
        Scoring method options include:

        - **membership**: number of instances belonging to each cluster

    legend : bool, default: True
        Whether or not to draw the size legend onto the graph, omit the legend
        to more easily see clusters that overlap.

    legend_loc : str, default: "lower left"
        The location of the legend on the graph, used to move the legend out
        of the way of clusters into open space. The same legend location
        options for matplotlib are used here.

        .. seealso:: https://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.legend

    legend_size : float, default: 1.5
        The size, in inches, of the size legend to inset into the graph.

    random_state : int or RandomState, default: None
        Fixes the random state for stochastic embedding algorithms.

    is_fitted : bool or str, default='auto'
        Specify if the wrapped estimator is already fitted. If False, the estimator
        will be fit when the visualizer is fit, otherwise, the estimator will not be
        modified. If 'auto' (default), a helper method will check if the estimator
        is fitted before fitting it again.

    kwargs : dict
        Keyword arguments passed to the base class and may influence the
        feature visualization properties.

    Attributes
    ----------
    cluster_centers_ : array of shape (n_clusters, n_features)
        The computed cluster centers from the underlying model.

    embedded_centers_ : array of shape (n_clusters, 2)
        The positions of all the cluster centers on the graph.

    scores_ : array of shape (n_clusters,)
        The scores of each cluster that determine its size on the graph.

    fit_time_ : Timer
        The time it took to fit the clustering model and perform the embedding.

    Notes
    -----
    Currently the only two embeddings supported are MDS and TSNE. Soon to
    follow will be PCoA and a customized version of PCoA for LDA. The only
    supported scoring metric is membership, but in the future, silhouette
    scores and cluster diameter will be added.

    In terms of algorithm support, right now any clustering algorithm that has
    a learned ``cluster_centers_`` and ``labels_`` attribute will work with
    the visualizer. In the future, we will update this to work with hierarchical
    clusterers that have ``n_components`` and LDA.
    """

    def __init__(
        self,
        estimator,
        ax=None,
        min_size=400,
        max_size=25000,
        embedding="mds",
        scoring="membership",
        legend=True,
        legend_loc="lower left",
        legend_size=1.5,
        random_state=None,
        is_fitted="auto",
        **kwargs
    ):
        # Initialize the visualizer bases
        super(InterclusterDistance, self).__init__(estimator, ax=ax, **kwargs)

        # Ensure that a valid embedding and scoring is passed in
        validate_embedding(embedding)
        validate_scoring(scoring)

        # Set decomposition properties
        self.scoring = scoring
        self.embedding = embedding
        self.random_state = random_state

        # Set visual properties
        self.legend = legend
        self.min_size = min_size
        self.max_size = max_size
        self.legend_loc = legend_loc
        self.legend_size = legend_size

        # Colors are currently hardcoded, need to compute face and edge color
        # from this color based on the alpha of the cluster center. The user
        # can "hack" these properties before drawing, however.
        self.facecolor = "#2e719344"
        self.edgecolor = "#2e719399"

        if self.legend:
            self.lax  # If legend True, test the version availability

    @memoized
    def lax(self):
        """
        Returns the legend axes, creating it only on demand by creating a 2"
        by 2" inset axes that has no grid, ticks, spines or face frame (e.g
        is mostly invisible). The legend can then be drawn on this axes.
        """
        if inset_locator is None:
            raise YellowbrickValueError(
                (
                    "intercluster distance map legend requires matplotlib 2.0.2 or "
                    "later please upgrade matplotlib or set legend=False "
                )
            )

        lax = inset_locator.inset_axes(
            self.ax,
            width=self.legend_size,
            height=self.legend_size,
            loc=self.legend_loc,
        )

        lax.set_frame_on(False)
        lax.set_facecolor("none")
        lax.grid(False)
        lax.set_xlim(-1.4, 1.4)
        lax.set_ylim(-1.4, 1.4)
        lax.set_xticks([])
        lax.set_yticks([])

        for name in lax.spines:
            lax.spines[name].set_visible(False)

        return lax

    @memoized
    def transformer(self):
        """
        Creates the internal transformer that maps the cluster center's high
        dimensional space to its two dimensional space.
        """
        ttype = self.embedding.lower()  # transformer method type

        if ttype == "mds":
            return MDS(n_components=2, random_state=self.random_state)

        if ttype == "tsne":
            return TSNE(n_components=2, random_state=self.random_state)

        raise YellowbrickValueError("unknown embedding '{}'".format(ttype))

    @property
    def cluster_centers_(self):
        """
        Searches for or creates cluster centers for the specified clustering
        algorithm. This algorithm ensures that that the centers are
        appropriately drawn and scaled so that distance between clusters are
        maintained.
        """
        # TODO: Handle agglomerative clustering and LDA
        for attr in ("cluster_centers_",):
            try:
                return getattr(self.estimator, attr)
            except AttributeError:
                continue

        raise AttributeError(
            "could not find or make cluster_centers_ for {}".format(
                self.estimator.__class__.__name__
            )
        )

    def fit(self, X, y=None):
        """
        Fit the clustering model, computing the centers then embeds the centers
        into 2D space using the embedding method specified.
        """
        with Timer() as self.fit_time_:
            if not check_fitted(self.estimator, is_fitted_by=self.is_fitted):
                # Fit the underlying estimator
                self.estimator.fit(X, y)

        # Get the centers
        # TODO: is this how sklearn stores all centers in the model?
        C = self.cluster_centers_

        # Embed the centers in 2D space and get the cluster scores
        self.embedded_centers_ = self.transformer.fit_transform(C)
        self.scores_ = self._score_clusters(X, y)

        # Draw the clusters and fit returns self
        self.draw()
        return self

    def draw(self):
        """
        Draw the embedded centers with their sizes on the visualization.
        """
        # Compute the sizes of the markers from their score
        sizes = self._get_cluster_sizes()

        # Draw the scatter plots with associated sizes on the graph
        self.ax.scatter(
            self.embedded_centers_[:, 0],
            self.embedded_centers_[:, 1],
            s=sizes,
            c=self.facecolor,
            edgecolor=self.edgecolor,
            linewidth=1,
        )

        # Annotate the clusters with their labels
        for i, pt in enumerate(self.embedded_centers_):
            self.ax.text(
                s=str(i), x=pt[0], y=pt[1], va="center", ha="center", fontweight="bold"
            )

        # Ensure the current axes is always the main residuals axes
        plt.sca(self.ax)
        return self.ax

    def finalize(self):
        """
        Finalize the visualization to create an "origin grid" feel instead of
        the default matplotlib feel. Set the title, remove spines, and label
        the grid with components. This function also adds a legend from the
        sizes if required.
        """
        # Set the default title if a user hasn't supplied one
        self.set_title(
            "{} Intercluster Distance Map (via {})".format(
                self.estimator.__class__.__name__, self.embedding.upper()
            )
        )

        # Create the origin grid and minimalist display
        self.ax.set_xticks([0])
        self.ax.set_yticks([0])
        self.ax.set_xticklabels([])
        self.ax.set_yticklabels([])
        self.ax.set_xlabel("PC2")
        self.ax.set_ylabel("PC1")

        # Make the legend by creating an inset axes that shows relative sizing
        # based on the scoring metric supplied by the user.
        if self.legend:
            self._make_size_legend()

    def _score_clusters(self, X, y=None):
        """
        Determines the "scores" of the cluster, the metric that determines the
        size of the cluster visualized on the visualization.
        """
        stype = self.scoring.lower()  # scoring method name

        if stype == "membership":
            return np.bincount(self.estimator.labels_)

        raise YellowbrickValueError("unknown scoring method '{}'".format(stype))

    def _get_cluster_sizes(self):
        """
        Returns the marker size (in points, e.g. area of the circle) based on
        the scores, using the prop_to_size scaling mechanism.
        """
        # NOTE: log and power are hardcoded, should we allow the user to specify?
        return prop_to_size(
            self.scores_, mi=self.min_size, ma=self.max_size, log=False, power=0.5
        )

    def _make_size_legend(self):
        """
        Draw a legend that shows relative sizes of the clusters at the 25th,
        50th, and 75th percentile based on the current scoring metric.
        """
        # Compute the size of the markers and scale them to our figure size
        # NOTE: the marker size is the area of the plot, we need to compute the
        # radius of the markers.
        areas = self._get_cluster_sizes()
        radii = np.sqrt(areas / np.pi)
        scaled = np.interp(radii, (radii.min(), radii.max()), (0.1, 1))

        # Compute the locations of the 25th, 50th, and 75th percentile scores
        indices = np.array([percentile_index(self.scores_, p) for p in (25, 50, 75)])

        # Draw size circles annotated with the percentile score as the legend.
        for idx in indices:
            # TODO: should the size circle's center be hard coded like this?
            center = (-0.30, 1 - scaled[idx])
            c = Circle(
                center,
                scaled[idx],
                facecolor="none",
                edgecolor="#2e7193",
                linewidth=1.5,
                linestyle="--",
            )
            self.lax.add_patch(c)

            # Add annotation to the size circle with the value of the score
            self.lax.annotate(
                self.scores_[idx],
                (-0.30, 1 - (2 * scaled[idx])),
                xytext=(1, 1 - (2 * scaled[idx])),
                arrowprops=dict(arrowstyle="wedge", color="#2e7193"),
                va="center",
                ha="center",
            )

        # Draw size legend title
        self.lax.text(s="membership", x=0, y=1.2, va="center", ha="center")

        # Ensure the current axes is always the main axes after modifying the
        # inset axes and while drawing.
        plt.sca(self.ax)


# alias
ICDM = InterclusterDistance

##########################################################################
## Helper Methods
##########################################################################


def percentile_index(a, q):
    """
    Returns the index of the value at the Qth percentile in array a.
    """
    return np.where(a == np.percentile(a, q, interpolation="nearest"))[0][0]


def validate_string_param(s, valid, param_name="param"):
    """
    Raises a well formatted exception if s is not in valid, otherwise does not
    raise an exception. Uses ``param_name`` to identify the parameter.
    """
    if s.lower() not in valid:
        raise YellowbrickValueError(
            "unknown {} '{}', chose from '{}'".format(param_name, s, ", ".join(valid))
        )


def validate_embedding(param):
    """
    Raises an exception if the param is not in VALID_EMBEDDING
    """
    validate_string_param(param, VALID_EMBEDDING, "embedding")


def validate_scoring(param):
    """
    Raises an exception if the param is not in VALID_SCORING
    """
    validate_string_param(param, VALID_SCORING, "scoring")


##########################################################################
## Quick Method
##########################################################################


def intercluster_distance(
    estimator,
    X,
    y=None,
    ax=None,
    min_size=400,
    max_size=25000,
    embedding="mds",
    scoring="membership",
    legend=True,
    legend_loc="lower left",
    legend_size=1.5,
    random_state=None,
    is_fitted="auto",
    show=True,
    **kwargs
):
    """Quick Method:
    Intercluster distance maps display an embedding of the cluster centers in
    2 dimensions with the distance to other centers preserved. E.g. the closer
    to centers are in the visualization, the closer they are in the original
    feature space. The clusters are sized according to a scoring metric. By
    default, they are sized by membership, e.g. the number of instances that
    belong to each center. This gives a sense of the relative importance of
    clusters. Note however, that because two clusters overlap in the 2D space,
    it does not imply that they overlap in the original feature space.

    Parameters
    ----------
    estimator : a Scikit-Learn clusterer
        Should be an instance of a centroidal clustering algorithm (or a
        hierarchical algorithm with a specified number of clusters). Also
        accepts some other models like LDA for text clustering.
        If it is not a clusterer, an exception is raised. If the estimator
        is not fitted, it is fit when the visualizer is fitted, unless
        otherwise specified by ``is_fitted``.

    X : array-like of shape (n, m)
        A matrix or data frame with n instances and m features

    y : array-like of shape (n,), optional
        A vector or series representing the target for each instance

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    min_size : int, default: 400
        The size, in points, of the smallest cluster drawn on the graph.
        Cluster sizes will be scaled between the min and max sizes.

    max_size : int, default: 25000
        The size, in points, of the largest cluster drawn on the graph.
        Cluster sizes will be scaled between the min and max sizes.

    embedding : default: 'mds'
        The algorithm used to embed the cluster centers in 2 dimensional space
        so that the distance between clusters is represented equivalently to
        their relationship in feature spaceself.
        Embedding algorithm options include:

        - **mds**: multidimensional scaling
        - **tsne**: stochastic neighbor embedding

    scoring : default: 'membership'
        The scoring method used to determine the size of the clusters drawn on
        the graph so that the relative importance of clusters can be viewed.
        Scoring method options include:

        - **membership**: number of instances belonging to each cluster

    legend : bool, default: True
        Whether or not to draw the size legend onto the graph, omit the legend
        to more easily see clusters that overlap.

    legend_loc : str, default: "lower left"
        The location of the legend on the graph, used to move the legend out
        of the way of clusters into open space. The same legend location
        options for matplotlib are used here.

        .. seealso:: https://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.legend

    legend_size : float, default: 1.5
        The size, in inches, of the size legend to inset into the graph.

    random_state : int or RandomState, default: None
        Fixes the random state for stochastic embedding algorithms.

    is_fitted : bool or str, default='auto'
        Specify if the wrapped estimator is already fitted. If False, the estimator
        will be fit when the visualizer is fit, otherwise, the estimator will not be
        modified. If 'auto' (default), a helper method will check if the estimator
        is fitted before fitting it again.

    show : bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however
        you cannot call ``plt.savefig`` from this signature, nor
        ``clear_figure``. If False, simply calls ``finalize()``

    kwargs : dict
        Keyword arguments passed to the base class and may influence the
        feature visualization properties.

    Returns
    -------
    viz : InterclusterDistance
        The intercluster distance visualizer, fitted and finalized.
    """
    oz = InterclusterDistance(
        estimator,
        ax=ax,
        min_size=min_size,
        max_size=max_size,
        embedding=embedding,
        scoring=scoring,
        legend=legend,
        legend_loc=legend_loc,
        legend_size=legend_size,
        random_state=random_state,
        is_fitted=is_fitted,
        **kwargs
    )

    oz.fit(X, y)

    if show:
        oz.show()
    else:
        oz.finalize()

    return oz



# yellowbrick.features.pcoords
# Implementations of parallel coordinates for feature analysis.
#
# Author:  Benjamin Bengfort
# Author:  @thekylesaurus
# Created: Mon Oct 03 21:46:06 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: pcoords.py [0f4b236] benjamin@bengfort.com $

"""
Implementation of parallel coordinates for multi-dimensional feature analysis.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np

from numpy.random import RandomState
from sklearn.preprocessing import MinMaxScaler, MaxAbsScaler
from sklearn.preprocessing import Normalizer, StandardScaler

from yellowbrick.draw import manual_legend
from yellowbrick.features.base import DataVisualizer
from yellowbrick.utils import is_dataframe, is_series
from yellowbrick.exceptions import YellowbrickTypeError, YellowbrickValueError


##########################################################################
## Quick Methods
##########################################################################

def parallel_coordinates(
    X,
    y,
    ax=None,
    features=None,
    classes=None,
    normalize=None,
    sample=1.0,
    random_state=None,
    shuffle=False,
    colors=None,
    colormap=None,
    alpha=None,
    fast=False,
    vlines=True,
    vlines_kwds=None,
    show=True,
    **kwargs
):
    """Displays each feature as a vertical axis and each instance as a line.

    This helper function is a quick wrapper to utilize the ParallelCoordinates
    Visualizer (Transformer) for one-off analysis.

    Parameters
    ----------
    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features

    y : ndarray or Series of length n
        An array or series of target or class values

    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    features : list, default: None
        a list of feature names to use
        If a DataFrame is passed to fit and features is None, feature
        names are selected as the columns of the DataFrame.

    classes : list, default: None
        a list of class names for the legend
        If classes is None and a y value is passed to fit then the classes
        are selected from the target vector.

    normalize : string or None, default: None
        specifies which normalization method to use, if any
        Current supported options are 'minmax', 'maxabs', 'standard', 'l1',
        and 'l2'.

    sample : float or int, default: 1.0
        specifies how many examples to display from the data
        If int, specifies the maximum number of samples to display.
        If float, specifies a fraction between 0 and 1 to display.

    random_state : int, RandomState instance or None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by np.random; only used if shuffle is True and sample < 1.0

    shuffle : boolean, default: True
        specifies whether sample is drawn randomly

    colors : list or tuple, default: None
        optional list or tuple of colors to colorize lines
        Use either color to colorize the lines on a per class basis or
        colormap to color them on a continuous scale.

    colormap : string or cmap, default: None
        optional string or matplotlib cmap to colorize lines
        Use either color to colorize the lines on a per class basis or
        colormap to color them on a continuous scale.

    alpha : float, default: None
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered lines more visible.
        If None, the alpha is set to 0.5 in "fast" mode and 0.25 otherwise.

    fast : bool, default: False
        Fast mode improves the performance of the drawing time of parallel
        coordinates but produces an image that does not show the overlap of
        instances in the same class. Fast mode should be used when drawing all
        instances is too burdensome and sampling is not an option.

    vlines : boolean, default: True
        flag to determine vertical line display

    vlines_kwds : dict, default: None
        options to style or display the vertical lines, default: None

    show : bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Returns
    -------
    viz : ParallelCoordinates
        Returns the fitted, finalized visualizer
    """
    # Instantiate the visualizer
    visualizer = ParallelCoordinates(
        ax=ax,
        features=features,
        classes=classes,
        normalize=normalize,
        sample=sample,
        random_state=random_state,
        shuffle=shuffle,
        colors=colors,
        colormap=colormap,
        alpha=alpha,
        fast=fast,
        vlines=vlines,
        vlines_kwds=vlines_kwds,
        **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(X, y, **kwargs)
    visualizer.transform(X)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer object
    return visualizer


##########################################################################
## Static Parallel Coordinates Visualizer
##########################################################################


class ParallelCoordinates(DataVisualizer):
    """
    Parallel coordinates displays each feature as a vertical axis spaced
    evenly along the horizontal, and each instance as a line drawn between
    each individual axis. This allows you to detect braids of similar instances
    and separability that suggests a good classification problem.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    features : list, default: None
        a list of feature names to use
        If a DataFrame is passed to fit and features is None, feature
        names are selected as the columns of the DataFrame.

    classes : list, default: None
        a list of class names for the legend
        The class labels for each class in y, ordered by sorted class index. These
        names act as a label encoder for the legend, identifying integer classes
        or renaming string labels. If omitted, the class labels will be taken from
        the unique values in y.

        Note that the length of this list must match the number of unique values in
        y, otherwise an exception is raised.

    normalize : string or None, default: None
        specifies which normalization method to use, if any
        Current supported options are 'minmax', 'maxabs', 'standard', 'l1',
        and 'l2'.

    sample : float or int, default: 1.0
        specifies how many examples to display from the data
        If int, specifies the maximum number of samples to display.
        If float, specifies a fraction between 0 and 1 to display.

    random_state : int, RandomState instance or None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by np.random; only used if shuffle is True and sample < 1.0

    shuffle : boolean, default: True
        specifies whether sample is drawn randomly

    colors : list or tuple, default: None
        A single color to plot all instances as or a list of colors to color each
        instance according to its class. If not enough colors per class are
        specified then the colors are treated as a cycle.

    colormap : string or cmap, default: None
        The colormap used to create the individual colors. If classes are
        specified the colormap is used to evenly space colors across each class.

    alpha : float, default: None
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered lines more visible.
        If None, the alpha is set to 0.5 in "fast" mode and 0.25 otherwise.

    fast : bool, default: False
        Fast mode improves the performance of the drawing time of parallel
        coordinates but produces an image that does not show the overlap of
        instances in the same class. Fast mode should be used when drawing all
        instances is too burdensome and sampling is not an option.

    vlines : boolean, default: True
        flag to determine vertical line display

    vlines_kwds : dict, default: None
        options to style or display the vertical lines, default: None

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    n_samples_ : int
        number of samples included in the visualization object

    features_ : ndarray, shape (n_features,)
        The names of the features discovered or used in the visualizer that
        can be used as an index to access or modify data in X. If a user passes
        feature names in, those features are used. Otherwise the columns of a
        DataFrame are used or just simply the indices of the data array.

    classes_ : ndarray, shape (n_classes,)
        The class labels that define the discrete values in the target. Only
        available if the target type is discrete. This is guaranteed to be
        strings even if the classes are a different type.

    Examples
    --------

    >>> visualizer = ParallelCoordinates()
    >>> visualizer.fit(X, y)
    >>> visualizer.transform(X)
    >>> visualizer.show()
    """

    NORMALIZERS = {
        "minmax": MinMaxScaler(),
        "maxabs": MaxAbsScaler(),
        "standard": StandardScaler(),
        "l1": Normalizer("l1"),
        "l2": Normalizer("l2"),
    }

    def __init__(
        self,
        ax=None,
        features=None,
        classes=None,
        normalize=None,
        sample=1.0,
        random_state=None,
        shuffle=False,
        colors=None,
        colormap=None,
        alpha=None,
        fast=False,
        vlines=True,
        vlines_kwds=None,
        **kwargs
    ):
        if "target_type" not in kwargs:
            kwargs["target_type"] = "discrete"
        super(ParallelCoordinates, self).__init__(
            ax=ax,
            features=features,
            classes=classes,
            colors=colors,
            colormap=colormap,
            **kwargs
        )

        # Validate 'normalize' argument
        if normalize in self.NORMALIZERS or normalize is None:
            self.normalize = normalize
        else:
            raise YellowbrickValueError(
                "'{}' is an unrecognized normalization method".format(normalize)
            )

        # Validate 'sample' argument
        if isinstance(sample, int):
            if sample < 1:
                raise YellowbrickValueError(
                    "`sample` parameter of type `int` must be greater than 1"
                )
        elif isinstance(sample, float):
            if sample <= 0 or sample > 1:
                raise YellowbrickValueError(
                    "`sample` parameter of type `float` must be between 0 and 1"
                )
        else:
            raise YellowbrickTypeError("`sample` parameter must be int or float")
        self.sample = sample

        # Set sample parameters
        if isinstance(shuffle, bool):
            self.shuffle = shuffle
        else:
            raise YellowbrickTypeError("`shuffle` parameter must be boolean")
        if self.shuffle:
            if (random_state is None) or isinstance(random_state, int):
                self._rng = RandomState(random_state)
            elif isinstance(random_state, RandomState):
                self._rng = random_state
            else:
                raise YellowbrickTypeError(
                    "`random_state` must be None, int, or np.random.RandomState"
                )
        else:
            self._rng = None

        # Visual and drawing parameters
        self.fast = fast
        self.alpha = alpha
        self.show_vlines = vlines
        self.vlines_kwds = vlines_kwds or {"linewidth": 1, "color": "black"}

        # Internal properties
        self._increments = None
        self._colors = None

    def fit(self, X, y=None, **kwargs):
        """
        The fit method is the primary drawing input for the
        visualization since it has both the X and y data required for the
        viz and the transform method does not.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features

        y : ndarray or Series of length n
            An array or series of target or class values

        kwargs : dict
            Pass generic arguments to the drawing method

        Returns
        -------
        self : instance
            Returns the instance of the transformer/visualizer
        """
        # Determine the features, classes, and colors
        super(ParallelCoordinates, self).fit(X, y)

        # Convert from pandas data types
        if is_dataframe(X):
            X = X.values
        if is_series(y):
            y = y.values

        # Ticks for each feature specified
        self._increments = np.arange(len(self.features_))

        # Subsample instances
        X, y = self._subsample(X, y)

        # Normalize instances
        if self.normalize is not None:
            X = self.NORMALIZERS[self.normalize].fit_transform(X)

        self.draw(X, y, **kwargs)
        return self

    def draw(self, X, y, **kwargs):
        """
        Called from the fit method, this method creates the parallel
        coordinates canvas and draws each instance and vertical lines on it.

        Parameters
        ----------
        X : ndarray of shape n x m
            A matrix of n instances with m features

        y : ndarray of length n
            An array or series of target or class values

        kwargs : dict
            Pass generic arguments to the drawing method

        """
        if self.fast:
            return self.draw_classes(X, y, **kwargs)
        return self.draw_instances(X, y, **kwargs)

    def draw_instances(self, X, y, **kwargs):
        """
        Draw the instances colored by the target y such that each line is a
        single instance. This is the "slow" mode of drawing, since each
        instance has to be drawn individually. However, in so doing, the
        density of instances in braids is more apparent since lines have an
        independent alpha that is compounded in the figure.

        This is the default method of drawing.

        Parameters
        ----------
        X : ndarray of shape n x m
            A matrix of n instances with m features

        y : ndarray of length n
            An array or series of target or class values

        Notes
        -----
        This method can be used to draw additional instances onto the parallel
        coordinates before the figure is finalized.
        """
        # Get alpha from param or default
        alpha = self.alpha or 0.25

        for idx in range(len(X)):
            Xi = X[idx]
            yi = y[idx]
            color = self.get_colors([yi])[0]

            self.ax.plot(self._increments, Xi, color=color, alpha=alpha, **kwargs)

        return self.ax

    def draw_classes(self, X, y, **kwargs):
        """
        Draw the instances colored by the target y such that each line is a
        single class. This is the "fast" mode of drawing, since the number of
        lines drawn equals the number of classes, rather than the number of
        instances. However, this drawing method sacrifices inter-class density
        of points using the alpha parameter.

        Parameters
        ----------
        X : ndarray of shape n x m
            A matrix of n instances with m features

        y : ndarray of length n
            An array or series of target or class values
        """
        # Get alpha from param or default
        alpha = self.alpha or 0.5

        # Prepare to flatten data within each class:
        #   introduce separation between individual data points using None in
        #   x-values and arbitrary value (one) in y-values
        X_separated = np.hstack([X, np.ones((X.shape[0], 1))])
        increments_separated = self._increments.tolist()
        increments_separated.append(None)

        # Get the classes that exist in the dataset, y
        y_values = np.unique(y)

        # Plot each class as a single line plot
        for yi in y_values:
            color = self.get_colors([yi])[0]

            X_in_class = X_separated[y == yi, :]
            increments_in_class = increments_separated * len(X_in_class)
            if len(X_in_class) > 0:
                self.ax.plot(
                    increments_in_class,
                    X_in_class.flatten(),
                    linewidth=1,
                    color=color,
                    alpha=alpha,
                    **kwargs
                )

        return self.ax

    def finalize(self, **kwargs):
        """
        Performs the final rendering for the multi-axis visualization, including
        setting and rendering the vertical axes each instance is plotted on. Adds
        a title, a legend, and manages the grid.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        Notes
        -----
        Generally this method is called from show and not directly by the user.
        """
        # Set the title
        self.set_title(
            "Parallel Coordinates for {} Features".format(len(self.features_))
        )

        # Add the vertical lines
        # TODO: Make an independent function for override!
        if self.show_vlines:
            for idx in self._increments:
                self.ax.axvline(idx, **self.vlines_kwds)

        # Set the limits
        self.ax.set_xticks(self._increments)
        self.ax.set_xticklabels(self.features_)
        self.ax.set_xlim(self._increments[0], self._increments[-1])

        # Add the legend sorting classes by name
        labels = sorted(list(self._colors.keys()))
        colors = [self._colors[lbl] for lbl in labels]
        manual_legend(self, labels, colors, loc="best", frameon=True)

        # Add the grid view
        self.ax.grid()

    def _subsample(self, X, y):

        # Choose a subset of samples
        if isinstance(self.sample, int):
            n_samples = min([self.sample, len(X)])
        elif isinstance(self.sample, float):
            n_samples = int(len(X) * self.sample)

        if (n_samples < len(X)) and self.shuffle:
            indices = self._rng.choice(len(X), n_samples, replace=False)
        else:
            indices = slice(n_samples)
        X = X[indices, :]
        y = y[indices]

        self.n_samples_ = n_samples
        return X, y



# yellowbrick.text.freqdist
# Implementations of frequency distributions for text visualization.
#
# Author:   Rebecca Bilbro
# Created:  Mon Feb 20 12:38:20 2017 -0500
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: freqdist.py [67b2740] rebecca.bilbro@bytecubed.com $

"""
Implementations of frequency distributions for text visualization
"""

##########################################################################
## Imports
##########################################################################

import numpy as np

from operator import itemgetter

from yellowbrick.text.base import TextVisualizer
from yellowbrick.exceptions import YellowbrickValueError


##########################################################################
## Quick Method
##########################################################################


def freqdist(
    features, X, y=None, ax=None, n=50, orient="h", color=None, show=True, **kwargs
):
    """Displays frequency distribution plot for text.

    This helper function is a quick wrapper to utilize the FreqDist
    Visualizer (Transformer) for one-off analysis.

    Parameters
    ----------

    features : list, default: None
        The list of feature names from the vectorizer, ordered by index. E.g.
        a lexicon that specifies the unique vocabulary of the corpus. This
        can be typically fetched using the ``get_feature_names()`` method of
        the transformer in Scikit-Learn.

    X: ndarray or DataFrame of shape n x m
        A matrix of n instances with m features. In the case of text,
        X is a list of list of already preprocessed words

    y: ndarray or Series of length n
        An array or series of target or class values

    ax : matplotlib axes, default: None
        The axes to plot the figure on.

    n: integer, default: 50
        Top N tokens to be plotted.

    orient : 'h' or 'v', default: 'h'
        Specifies a horizontal or vertical bar chart.

    color : string
        Specify color for bars

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however
        you cannot call ``plt.savefig`` from this signature, nor
        ``clear_figure``. If False, simply calls ``finalize()``

    kwargs: dict
        Keyword arguments passed to the super class.

    Returns
    -------
    visualizer: FreqDistVisualizer
        Returns the fitted, finalized visualizer
    """
    # Instantiate the visualizer
    viz = FreqDistVisualizer(features, ax=ax, n=n, orient=orient, color=color, **kwargs)

    # Fit and transform the visualizer (calls draw)
    viz.fit(X, y, **kwargs)
    viz.transform(X)

    # Draw the final visualization
    if show:
        viz.show()
    else:
        viz.finalize()

    # Return the visualizer object
    return viz


class FrequencyVisualizer(TextVisualizer):
    """
    A frequency distribution tells us the frequency of each vocabulary
    item in the text. In general, it could count any kind of observable
    event. It is a distribution because it tells us how the total
    number of word tokens in the text are distributed across the
    vocabulary items.


    Parameters
    ----------
    features : list, default: None
        The list of feature names from the vectorizer, ordered by index. E.g.
        a lexicon that specifies the unique vocabulary of the corpus. This
        can be typically fetched using the ``get_feature_names()`` method of
        the transformer in Scikit-Learn.

    ax : matplotlib axes, default: None
        The axes to plot the figure on.

    n: integer, default: 50
        Top N tokens to be plotted.

    orient : 'h' or 'v', default: 'h'
        Specifies a horizontal or vertical bar chart.

    color : string
        Specify color for bars

    kwargs : dict
        Pass any additional keyword arguments to the super class.

    These parameters can be influenced later on in the visualization
    process, but can and should be set as early as possible.
    """

    def __init__(self, features, ax=None, n=50, orient="h", color=None, **kwargs):
        super(FreqDistVisualizer, self).__init__(ax=ax, **kwargs)

        # Check that the orient is correct
        orient = orient.lower().strip()
        if orient not in {"h", "v"}:
            raise YellowbrickValueError("Orientation must be 'h' or 'v'")

        # Visualizer parameters
        self.N = n
        self.features = features

        # Visual arguments
        self.color = color
        self.orient = orient

    def count(self, X):
        """
        Called from the fit method, this method gets all the
        words from the corpus and their corresponding frequency
        counts.

        Parameters
        ----------

        X : ndarray or masked ndarray
            Pass in the matrix of vectorized documents, can be masked in
            order to sum the word frequencies for only a subset of documents.

        Returns
        -------

        counts : array
            A vector containing the counts of all words in X (columns)

        """
        # Sum on axis 0 (by columns), each column is a word
        # Convert the matrix to an array
        # Squeeze to remove the 1 dimension objects (like ravel)
        return np.squeeze(np.asarray(X.sum(axis=0)))

    def fit(self, X, y=None):
        """
        The fit method is the primary drawing input for the frequency
        distribution visualization. It requires vectorized lists of
        documents and a list of features, which are the actual words
        from the original corpus (needed to label the x-axis ticks).

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features representing the corpus
            of frequency vectorized documents.

        y : ndarray or DataFrame of shape n
            Labels for the documents for conditional frequency distribution.

        Notes
        -----
        .. note:: Text documents must be vectorized before ``fit()``.
        """

        # Compute the conditional word frequency
        if y is not None:
            # Fit the frequencies
            self.conditional_freqdist_ = {}

            # Conditional frequency distribution
            self.classes_ = [str(label) for label in set(y)]
            for label in self.classes_:
                self.conditional_freqdist_[label] = self.count(X[y == label])
        else:
            # No conditional frequencies
            self.conditional_freqdist_ = None

        # Frequency distribution of entire corpus.
        self.freqdist_ = self.count(X)
        self.sorted_ = self.freqdist_.argsort()[::-1]  # Descending order

        # Compute the number of words, vocab, and hapaxes
        self.vocab_ = self.freqdist_.shape[0]
        self.words_ = self.freqdist_.sum()
        self.hapaxes_ = sum(1 for c in self.freqdist_ if c == 1)

        # Draw and ensure that we return self
        self.draw()
        return self

    def draw(self, **kwargs):
        """
        Called from the fit method, this method creates the canvas and
        draws the distribution plot on it.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        """
        # Prepare the data
        bins = np.arange(self.N)
        words = [self.features[i] for i in self.sorted_[: self.N]]
        freqs = {}

        # Set up the bar plots
        if self.conditional_freqdist_:
            for label, values in sorted(
                self.conditional_freqdist_.items(), key=itemgetter(0)
            ):
                freqs[label] = [values[i] for i in self.sorted_[: self.N]]
        else:
            freqs["corpus"] = [self.freqdist_[i] for i in self.sorted_[: self.N]]

        # Draw a horizontal barplot
        if self.orient == "h":
            # Add the barchart, stacking if necessary
            for label, freq in freqs.items():
                self.ax.barh(bins, freq, label=label, color=self.color, align="center")

            # Set the y ticks to the words
            self.ax.set_yticks(bins)
            self.ax.set_yticklabels(words)

            # Order the features from top to bottom on the y axis
            self.ax.invert_yaxis()

            # Turn off y grid lines and turn on x grid lines
            self.ax.yaxis.grid(False)
            self.ax.xaxis.grid(True)

        # Draw a vertical barplot
        elif self.orient == "v":
            # Add the barchart, stacking if necessary
            for label, freq in freqs.items():
                self.ax.bar(bins, freq, label=label, color=self.color, align="edge")

            # Set the y ticks to the words
            self.ax.set_xticks(bins)
            self.ax.set_xticklabels(words, rotation=90)

            # Turn off x grid lines and turn on y grid lines
            self.ax.yaxis.grid(True)
            self.ax.xaxis.grid(False)

        # Unknown state
        else:
            raise YellowbrickValueError("Orientation must be 'h' or 'v'")

        return self.ax

    def finalize(self, **kwargs):
        """
        The finalize method executes any subclass-specific axes
        finalization steps. The user calls show & show calls finalize.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        """
        # Set the title
        self.set_title("Frequency Distribution of Top {} tokens".format(self.N))

        # Create the vocab, count, and hapaxes labels
        infolabel = "vocab: {:,}\nwords: {:,}\nhapax: {:,}".format(
            self.vocab_, self.words_, self.hapaxes_
        )

        self.ax.text(
            0.68,
            0.97,
            infolabel,
            transform=self.ax.transAxes,
            fontsize=9,
            verticalalignment="top",
            bbox={"boxstyle": "round", "facecolor": "white", "alpha": 0.8},
        )

        # Set the legend and the grid
        self.ax.legend(loc="upper right", frameon=True)


# Backwards compatibility alias
FreqDistVisualizer = FrequencyVisualizer



# yellowbrick.datasets.base
# Loading utilities for the yellowbrick datasets.
#
# Author:   Rebecca Bilbro
# Author:   Benjamin Bengfort
# Author:   Raul Peralta
# Created: Thu Jul 26 13:53:01 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: base.py [] benjamin@bengfort.com $

"""
Loading utilities for the yellowbrick datasets.
"""

##########################################################################
## Imports
##########################################################################

import os
import json
import numpy as np

from .download import download_data
from .path import find_dataset_path, dataset_exists

from yellowbrick.exceptions import DatasetsError
from yellowbrick.utils.decorators import memoized

try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
## Dataset Object
##########################################################################


class BaseDataset(object):
    """
    Base functionality for Dataset and Corpus objects.
    """

    def __init__(self, name, url=None, signature=None, data_home=None):
        self.url = url
        self.name = name
        self.data_home = data_home
        self.signature = signature

        # Check if the dataset exists, and if not - download it!
        if not dataset_exists(self.name, data_home=data_home):
            self.download()

    def download(self, replace=False):
        """
        Download the dataset from the hosted Yellowbrick data store and save
        it to the location specified by ``get_data_home``. The downloader
        verifies the download completed successfully and safely by comparing
        the expected signature with the SHA 256 signature of the downloaded
        archive file.

        Parameters
        ----------
        replace : bool, default: False
            If the data archive already exists, replace the dataset. If this is
            False and the dataset exists, an exception is raised.
        """
        download_data(
            self.url,
            self.signature,
            data_home=self.data_home,
            replace=replace,
            extract=True,
        )

    def contents(self):
        """
        Contents returns a list of the files in the data directory.
        """
        data = find_dataset_path(self.name, data_home=self.data_home, ext=None)
        return os.listdir(data)

    @memoized
    def README(self):
        """
        Returns the contents of the README.md file that describes the dataset
        in detail and contains attribution information.
        """
        path = find_dataset_path(self.name, data_home=self.data_home, fname="README.md")
        with open(path, "r") as f:
            return f.read()

    @memoized
    def meta(self):
        """
        Returns the contents of the meta.json file that describes important
        attributes about the dataset and modifies the behavior of the loader.
        """
        path = find_dataset_path(
            self.name, data_home=self.data_home, fname="meta.json", raises=False
        )
        if path is None:
            return None

        with open(path, "r") as f:
            return json.load(f)

    @memoized
    def citation(self):
        """
        Returns the contents of the citation.bib file that describes the source
        and provenance of the dataset or to cite for academic work.
        """
        path = find_dataset_path(
            self.name, data_home=self.data_home, fname="meta.json", raises=False
        )
        if path is None:
            return None

        with open(path, "r") as f:
            return f.read()


class Dataset(BaseDataset):
    """
    Datasets contain a reference to data on disk and provide utilities for
    quickly loading files and objects into a variety of formats. The most
    common use of the Dataset object is to load example datasets provided by
    Yellowbrick to run the examples in the documentation.

    The dataset by default will return the data as a numpy array, however if
    Pandas is installed, it is possible to access the data as a DataFrame and
    Series object. In either case, the data is represented by a features table,
    X and a target vector, y.

    Parameters
    ----------
    name : str
        The name of the dataset; should either be a folder in data home or
        specified in the yellowbrick.datasets.DATASETS variable. This name is
        used to perform all lookups and identify the dataset externally.

    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from YELLOWBRICK_DATA or the default returned by ``get_data_home``.

    url : str, optional
        The web location where the archive file of the dataset can be
        downloaded from.

    signature : str, optional
        The signature of the data archive file, used to verify that the latest
        version of the data has been downloaded and that the download hasn't
        been corrupted or modified in anyway.
    """

    def to_data(self):
        """
        Returns the data contained in the dataset as X and y where X is the
        features matrix and y is the target vector. If pandas is installed,
        the data will be returned as DataFrame and Series objects. Otherwise,
        the data will be returned as two numpy arrays.

        Returns
        -------
        X : array-like with shape (n_instances, n_features)
            A pandas DataFrame or numpy array describing the instance features.

        y : array-like with shape (n_instances,)
            A pandas Series or numpy array describing the target vector.
        """
        if pd is not None:
            return self.to_pandas()
        return self.to_numpy()

    def to_numpy(self):
        """
        Returns the dataset as two numpy arrays: X and y.

        Returns
        -------
        X : array-like with shape (n_instances, n_features)
            A numpy array describing the instance features.

        y : array-like with shape (n_instances,)
            A numpy array describing the target vector.
        """
        path = find_dataset_path(self.name, ext=".npz", data_home=self.data_home)
        with np.load(path, allow_pickle=False) as npf:
            if "X" not in npf or "y" not in npf:
                raise DatasetsError(
                    (
                        "the downloaded dataset was improperly packaged without numpy "
                        "arrays - please report this bug to the Yellowbrick maintainers!"
                    )
                )

            # TODO: How to handle the case where y is None?
            return npf["X"], npf["y"]

    def to_pandas(self):
        """
        Returns the dataset as two pandas objects: X and y.

        Returns
        -------
        X : DataFrame with shape (n_instances, n_features)
            A pandas DataFrame containing feature data and named columns.

        y : Series with shape (n_instances,)
            A pandas Series containing target data and an index that matches
            the feature DataFrame index.
        """
        # Ensure the metadata is valid before continuing
        if self.meta is None:
            raise DatasetsError(
                (
                    "the downloaded dataset was improperly packaged without meta.json "
                    "- please report this bug to the Yellowbrick maintainers!"
                )
            )

        if "features" not in self.meta or "target" not in self.meta:
            raise DatasetsError(
                (
                    "the downloaded dataset was improperly packaged without features "
                    "or target - please report this bug to the Yellowbrick maintainers!"
                )
            )

        # Load data frame and return features and target
        # TODO: Return y as None if there is no self.meta["target"]
        df = self.to_dataframe()
        return df[self.meta["features"]], df[self.meta["target"]]

    def to_dataframe(self):
        """
        Returns the entire dataset as a single pandas DataFrame.

        Returns
        -------
        df : DataFrame with shape (n_instances, n_columns)
            A pandas DataFrame containing the complete original data table
            including all targets (specified by the meta data) and all
            features (including those that might have been filtered out).
        """
        if pd is None:
            raise DatasetsError(
                "pandas is required to load DataFrame, it can be installed with pip"
            )

        path = find_dataset_path(self.name, ext=".csv.gz", data_home=self.data_home)
        return pd.read_csv(path, compression="gzip")


class Corpus(BaseDataset):
    """
    Corpus datasets contain a reference to documents on disk and provide
    utilities for quickly loading text data for use in machine learning
    workflows. The most common use of the corpus is to load the text analysis
    examples from the Yellowbrick documentation.

    Parameters
    ----------
    name : str
        The name of the corpus; should either be a folder in data home or
        specified in the yellowbrick.datasets.DATASETS variable. This name is
        used to perform all lookups and identify the corpus externally.

    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from YELLOWBRICK_DATA or the default returned by ``get_data_home``.

    url : str, optional
        The web location where the archive file of the corpus can be
        downloaded from.

    signature : str, optional
        The signature of the data archive file, used to verify that the latest
        version of the data has been downloaded and that the download hasn't
        been corrupted or modified in anyway.
    """

    @memoized
    def root(self):
        """
        Discovers and caches the root directory of the corpus.
        """
        return find_dataset_path(self.name, data_home=self.data_home, ext=None)

    @memoized
    def labels(self):
        """
        Return the unique labels assigned to the documents.
        """
        return [
            name
            for name in os.listdir(self.root)
            if os.path.isdir(os.path.join(self.root, name))
        ]

    @property
    def files(self):
        """
        Returns the list of file names for all documents.
        """
        return [
            os.path.join(self.root, label, name)
            for label in self.labels
            for name in os.listdir(os.path.join(self.root, label))
        ]

    @property
    def data(self):
        """
        Read all of the documents from disk into an in-memory list.
        """

        def read(path):
            with open(path, "r", encoding="UTF-8") as f:
                return f.read()

        return [read(f) for f in self.files]

    @property
    def target(self):
        """
        Returns the label associated with each item in data.
        """
        return [os.path.basename(os.path.dirname(f)) for f in self.files]



# yellowbrick.contrib.missing.dispersion
# Missing Values Dispersion Visualizer
#
# Author:  Nathan Danielsen
# Created: Fri Mar 29 5:17:36 2018 -0500
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: dispersion.py [1443e16] ndanielsen@users.noreply.github.com $

"""
Dispersion visualizer for locations of missing values by column against index position.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np

from yellowbrick.style.palettes import color_palette
from .base import MissingDataVisualizer


##########################################################################
## MissingValues Visualizer
##########################################################################


class MissingValuesDispersion(MissingDataVisualizer):
    """
    The Missing Values Dispersion visualizer shows the locations of missing (nan)
    values in the feature dataset by the order of the index.

    When y targets are supplied to fit, the output dispersion plot is color
    coded according to the target y that the element refers to.

    Parameters
    ----------
    alpha : float, default: 0.5
        A value for bending elments with the background.

    marker : matplotlib marker, default: |
        The marker used for each element coordinate in the plot

    classes : list, default: None
        A list of class names for the legend.
        If classes is None and a y value is passed to fit then the classes
        are selected from the target vector.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    features_ : np.array
        The feature labels ranked according to their importance

    classes_ : np.array
        The class labels for each of the target values

    Examples
    --------

    >>> from yellowbrick.contrib.missing import MissingValuesDispersion
    >>> visualizer = MissingValuesDispersion()
    >>> visualizer.fit(X, y=y)
    >>> visualizer.show()
    """

    def __init__(self, alpha=0.5, marker="|", classes=None, **kwargs):

        if "target_type" not in kwargs:
            kwargs["target_type"] = "single"
        super(MissingValuesDispersion, self).__init__(**kwargs)
        self.alpha = alpha
        self.marker = marker

        self.classes_ = classes

        # Convert to array if necessary to match estimator.classes_
        if self.classes_ is not None:
            self.classes_ = np.array(classes)

        # Set up classifier score visualization properties
        if self.classes_ is not None:
            n_colors = len(self.classes_)
        else:
            n_colors = None

        self.colors = color_palette(kwargs.pop("colors", None), n_colors)

    def get_nan_locs(self, **kwargs):
        """Gets the locations of nans in feature data and returns
        the coordinates in the matrix
        """
        if np.issubdtype(self.X.dtype, np.string_) or np.issubdtype(
            self.X.dtype, np.unicode_
        ):
            mask = np.where(self.X == "")
            nan_matrix = np.zeros(self.X.shape)
            nan_matrix[mask] = np.nan

        else:
            nan_matrix = self.X.astype(float)

        if self.y is None:
            return np.argwhere(np.isnan(nan_matrix))
        else:
            nan_locs = []
            for target_value in np.unique(self.y):
                indices = np.argwhere(self.y == target_value)
                target_matrix = nan_matrix[indices.flatten()]
                nan_target_locs = np.argwhere(np.isnan(target_matrix))
                nan_locs.append((target_value, nan_target_locs))

            return nan_locs

    def draw(self, X, y, **kwargs):
        """Called from the fit method, this method creates a scatter plot that
        draws each instance as a class or target colored point, whose location
        is determined by the feature data set.

        If y is not None, then it draws a scatter plot where each class is in a
        different color.
        """
        nan_locs = self.get_nan_locs()
        if y is None:
            x_, y_ = list(zip(*nan_locs))
            self.ax.scatter(x_, y_, alpha=self.alpha, marker=self.marker, label=None)
        else:
            self.draw_multi_dispersion_chart(nan_locs)

    def draw_multi_dispersion_chart(self, nan_locs):
        """Draws a multi dimensional dispersion chart, each color corresponds
        to a different target variable.
        """
        for index, nan_values in enumerate(nan_locs):
            label, nan_locations = nan_values

            # if features passed in then, label as such
            if self.classes_ is not None:
                label = self.classes_[index]

            color = self.colors[index]

            x_, y_ = list(zip(*nan_locations))
            self.ax.scatter(
                x_, y_, alpha=self.alpha, marker=self.marker, color=color, label=label
            )

    def finalize(self, **kwargs):
        """
        Sets the title and x-axis label and adds a legend. Also ensures that
        the y tick labels are set to the feature names.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        Notes
        -----
        Generally this method is called from show and not directly by the user.
        """
        # Set the title
        self.set_title("Dispersion of Missing Values by Feature")
        # the x locations for the groups
        tick_locations = np.arange(len(self.features_))

        self.ax.set_xlabel("Position by index")
        self.ax.set_yticks(tick_locations)
        self.ax.set_yticklabels(self.get_feature_names())
        self.ax.legend(loc="upper left", prop={"size": 5}, bbox_to_anchor=(1, 1))


##########################################################################
## Quick Method
##########################################################################


def missing_dispersion(
    X, y=None, ax=None, classes=None, alpha=0.5, marker="|", **kwargs
):
    """
    The Missing Values Dispersion visualizer shows the locations of missing (nan)
    values in the feature dataset by the order of the index.

    When y targets are supplied to fit, the output dispersion plot is color
    coded according to the target y that the element refers to.

    Parameters
    ----------
    alpha : float, default: 0.5
        A value for bending elments with the background.

    marker : matplotlib marker, default: |
        The marker used for each element coordinate in the plot

    classes : list, default: None
        A list of class names for the legend.
        If classes is None and a y value is passed to fit then the classes
        are selected from the target vector.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    features_ : np.array
        The feature labels ranked according to their importance

    classes_ : np.array
        The class labels for each of the target values

    Examples
    --------

    >>> from yellowbrick.contrib.missing import missing_dispersion
    >>> visualizer = missing_dispersion(X, y=y)

    """
    # Instantiate the visualizer
    visualizer = MissingValuesDispersion(
        ax=ax, classes=classes, alpha=alpha, marker=marker, **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(X, y)
    visualizer.show()

    # Return the axes object on the visualizer
    return visualizer.ax



# yellowbrick.text.postag
# Implementation of part-of-speech visualization for text.
#
# Author:   Rebecca Bilbro
# Created:  Sun Mar 5 18:07:06 2017 -0500
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: postag.py [849f5a8] rebecca.bilbro@bytecubed.com $

"""
Implementation of part-of-speech visualization for text,
enabling the user to visualize a single document or
small subset of documents.
"""

##########################################################################
# Imports
##########################################################################

import numpy as np
import importlib

from yellowbrick.draw import bar_stack
from yellowbrick.text.base import TextVisualizer
from yellowbrick.style.colors import resolve_colors
from yellowbrick.exceptions import YellowbrickValueError

##########################################################################
# Part-of-speech tag punctuation and labels
##########################################################################

# NOTE: Penn Treebank converts all sentence closers (!,?,;) to periods
PUNCT_TAGS = [".", ":", ",", "``", "''", "(", ")", "#", "$"]

TAGSET_NAMES = {"penn_treebank": "Penn Treebank", "universal": "Universal Dependencies"}

PENN_TAGS = [
    "noun",
    "verb",
    "adjective",
    "adverb",
    "preposition",
    "determiner",
    "pronoun",
    "conjunction",
    "infinitive",
    "wh- word",
    "modal",
    "possessive",
    "existential",
    "punctuation",
    "digit",
    "non-English",
    "interjection",
    "list",
    "symbol",
    "other",
]

UNIVERSAL_TAGS = [
    "noun",
    "verb",
    "adjective",
    "adverb",
    "adposition",
    "determiner",
    "pronoun",
    "conjunction",
    "infinitive",
    "punctuation",
    "number",
    "interjection",
    "symbol",
    "other",
]


##########################################################################
# PosTagVisualizer
##########################################################################


class PosTagVisualizer(TextVisualizer):
    """
    Parts of speech (e.g. verbs, nouns, prepositions, adjectives)
    indicate how a word is functioning within the context of a sentence.
    In English as in many other languages, a single word can function in
    multiple ways. Part-of-speech tagging lets us encode information not
    only about a word’s definition, but also its use in context (for
    example the words “ship” and “shop” can be either a verb or a noun,
    depending on the context).

    The PosTagVisualizer creates a bar chart to visualize the relative
    proportions of different parts-of-speech in a corpus.

    Note that the PosTagVisualizer requires documents to already be
    part-of-speech tagged; the visualizer expects the corpus to come in
    the form of a list of (document) lists of (sentence) lists of
    (tag, token) tuples.

    Parameters
    ----------
    ax : matplotlib axes
        The axes to plot the figure on.

    tagset: string
        The tagset that was used to perform part-of-speech tagging.
        Either "penn_treebank" or "universal", defaults to "penn_treebank".
        Use "universal" if corpus has been tagged using SpaCy.

    colors : list or tuple of strings
        Specify the colors for each individual part-of-speech. Will override
        colormap if both are provided.

    colormap : string or matplotlib cmap
        Specify a colormap to color the parts-of-speech.

    frequency: bool {True, False}, default: False
        If set to True, part-of-speech tags will be plotted according to frequency,
        from most to least frequent.

    stack : bool {True, False}, default : False
        Plot the PosTag frequency chart as a per-class stacked bar chart.
        Note that fit() requires y for this visualization.

    parser : string or None, default: None
        If set to a string, string must be in the form of 'parser_tagger' or 'parser'
        to use defaults (for spacy this is 'en_core_web_sm', for nltk this is 'word').
        The 'parser' argument is one of the accepted parsing libraries. Currently
        'nltk' and 'spacy' are the only accepted libraries. NLTK or SpaCy must be
        installed into your environment. 'tagger' is the tagset to use. For example
        'nltk_wordpunct' would use the NLTK library with 'wordpunct' tagset. Or
        'spacy_en_core_web_sm' would use SpaCy with the 'en_core_web_sm' tagset.

    kwargs : dict
        Pass any additional keyword arguments to the PosTagVisualizer.

    Attributes
    ----------
    pos_tag_counts_: dict
        Mapping of part-of-speech tags to counts.

    Examples
    --------
    >>> viz = PosTagVisualizer()
    >>> viz.fit(X)
    >>> viz.show()
    """

    def __init__(
        self,
        ax=None,
        tagset="penn_treebank",
        colormap=None,
        colors=None,
        frequency=False,
        stack=False,
        parser=None,
        **kwargs,
    ):
        super(PosTagVisualizer, self).__init__(ax=ax, **kwargs)

        self.tagset_names = TAGSET_NAMES

        if tagset not in self.tagset_names:
            raise YellowbrickValueError(
                "'{}' is an invalid tagset. Please choose one of {}.".format(
                    tagset, ", ".join(self.tagset_names.keys())
                )
            )
        else:
            self.tagset = tagset

        self.punct_tags = frozenset(PUNCT_TAGS)
        self.frequency = frequency
        self.colormap = colormap
        self.colors = colors
        self.stack = stack
        self.parser = parser

    @property
    def parser(self):
        return self._parser

    @parser.setter
    def parser(self, parser):
        accepted_parsers = ("nltk", "spacy")
        if not parser:
            self._parser = None
        elif parser.startswith(accepted_parsers):
            parser_tagger = parser.split("_", 1)  # split on the first occurrence of _

            parser_name = None
            tagger_name = None

            if len(parser_tagger) == 1:  # if only 'nltk' or 'spacy' is provided
                parser_name = parser_tagger[0]
            if len(parser_tagger) == 2:
                parser_name = parser_tagger[0]
                tagger_name = parser_tagger[1]

            try:
                importlib.import_module(parser_name)
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "Can't find module '{}' in this environment.".format(parser)
                )

            if parser_name == "nltk":
                nltk = importlib.import_module("nltk")
                try:
                    nltk.data.find("corpora/treebank")
                except LookupError:
                    raise LookupError(
                        "Error occured because nltk postag data is not available"
                    )

                nltk_taggers = ["word", "wordpunct"]

                if not tagger_name:
                    tagger_name = "word"
                    parser = parser_name + "_" + tagger_name
                if tagger_name not in nltk_taggers:
                    raise ValueError(
                        "If using NLTK, tagger should either be 'word' (default) or"
                        " 'wordpunct'."
                    )

            elif parser_name == "spacy":
                if not tagger_name:
                    tagger_name = "en_core_web_sm"
                    parser = parser_name + "_" + tagger_name
                try:
                    spacy = importlib.import_module("spacy")
                    spacy.load(tagger_name)
                except OSError:
                    raise OSError(
                        "Spacy model '{}' has not been downloaded into this"
                        " environment.".format(tagger_name)
                    )
            self._parser = parser
        else:
            raise ValueError(
                "{} is an invalid parser. Currently the supported parsers are 'nltk'"
                "and 'spacy'".format(parser)
            )

    def fit(self, X, y=None, **kwargs):
        """
        Fits the corpus to the appropriate tag map.
        Text documents must be tokenized & tagged before passing to fit
        if the 'parse' argument has not been specified at initialization.
        Otherwise X can be a raw text ready to be parsed.

        Parameters
        ----------
        X : list or generator or str (raw text)
            Should be provided as a list of documents or a generator
            that yields a list of documents that contain a list of
            sentences that contain (token, tag) tuples. If X is a
            string, the 'parse' argument should be specified as 'nltk'
            or 'spacy' in order to parse the raw documents.

        y : ndarray or Series of length n
            An optional array of target values that are  ignored by the
            visualizer.

        kwargs : dict
            Pass generic arguments to the drawing method

        Returns
        -------
        self : instance
            Returns the instance of the transformer/visualizer
        """
        self.labels_ = ["documents"]

        if self.parser:
            parser_name = self.parser.split("_", 1)[0]
            parse_func = getattr(self, "parse_{}".format(parser_name))
            X = parse_func(X)

        if self.stack:
            if y is None:
                raise YellowbrickValueError("Specify y for stack=True")
            self.labels_ = np.unique(y)

        if self.tagset == "penn_treebank":
            self.pos_tag_counts_ = self._penn_tag_map()
            self._handle_treebank(X, y)

        elif self.tagset == "universal":
            self.pos_tag_counts_ = self._uni_tag_map()
            self._handle_universal(X, y)

        self.draw()

        return self

    def parse_nltk(self, X):
        """
        Tag a corpora using NLTK tagging (Penn-Treebank) to produce a generator of
        tagged documents in the form of a list of (document) lists of (sentence)
        lists of (token, tag) tuples.

        Parameters
        ----------
        X : str (raw text) or list of paragraphs (containing str)

        """
        nltk = importlib.import_module("nltk")
        nltk.data.find("corpora/treebank")
        tagger = self.parser.split("_", 1)[1]

        if tagger == "word":
            for doc in X:
                yield [
                    nltk.pos_tag(nltk.word_tokenize(sent))
                    for sent in nltk.sent_tokenize(doc)
                ]
        elif tagger == "wordpunct":
            for doc in X:
                yield [
                    nltk.pos_tag(nltk.wordpunct_tokenize(sent))
                    for sent in nltk.sent_tokenize(doc)
                ]

    def parse_spacy(self, X):
        """
        Tag a corpora using SpaCy tagging (Universal Dependencies) to produce a
        generator of tagged documents in the form of a list of (document)
        lists of (sentence) lists of (token, tag) tuples.

        Parameters
        ----------
        X : str (raw text) or list of paragraphs (containing str)

        """
        spacy = importlib.import_module("spacy")
        tagger = self.parser.split("_", 1)[1]
        nlp = spacy.load(tagger)

        if isinstance(X, list):
            for doc in X:
                tagged = nlp(doc)
                yield [
                    [(token.text, token.pos_) for token in sents]
                    for sents in tagged.sents
                ]
        elif isinstance(X, str):
            tagged = nlp(X)
            yield [
                [(token.text, token.pos_) for token in sents] for sents in tagged.sents
            ]

    def _penn_tag_map(self):
        """
        Returns a Penn Treebank part-of-speech tag map.
        """
        self._pos_tags = PENN_TAGS
        return self._make_tag_map(PENN_TAGS)

    def _uni_tag_map(self):
        """
        Returns a Universal Dependencies part-of-speech tag map.
        """
        self._pos_tags = UNIVERSAL_TAGS
        return self._make_tag_map(UNIVERSAL_TAGS)

    def _make_tag_map(self, tagset):
        """
        Returns a map of the tagset to a counter unless stack=True then returns
        a map of labels to a map of tagset to counters.
        """
        # ensures the dict contains a zero counter per tag
        zeros = [0] * len(tagset)
        return {label: dict(zip(tagset, zeros)) for label in self.labels_}
        return dict(zip(tagset, zeros))

    def _handle_universal(self, X, y=None):
        """
        Scan through the corpus to compute counts of each Universal
        Dependencies part-of-speech.

        Parameters
        ----------
        X : list or generator
            Should be provided as a list of documents or a generator
            that yields a list of documents that contain a list of
            sentences that contain (token, tag) tuples.
        """
        jump = {
            # combine proper and regular nouns
            "NOUN": "noun",
            "PROPN": "noun",
            "ADJ": "adjective",
            "VERB": "verb",
            # include particles with adverbs
            "ADV": "adverb",
            "PART": "adverb",
            "ADP": "adposition",
            "PRON": "pronoun",
            "CCONJ": "conjunction",
            "PUNCT": "punctuation",
            "DET": "determiner",
            "NUM": "number",
            "INTJ": "interjection",
            "SYM": "symbol",
        }

        for idx, tagged_doc in enumerate(X):
            for tagged_sent in tagged_doc:
                for _, tag in tagged_sent:
                    if tag == "SPACE":
                        continue
                    if self.stack:
                        counter = self.pos_tag_counts_[y[idx]]
                    else:
                        counter = self.pos_tag_counts_["documents"]

                    counter[jump.get(tag, "other")] += 1

    def _handle_treebank(self, X, y=None):
        """
        Create a part-of-speech tag mapping using the Penn Treebank tags

        Parameters
        ----------
        X : list or generator
            Should be provided as a list of documents or a generator
            that yields a list of documents that contain a list of
            sentences that contain (token, tag) tuples.
        """
        for idx, tagged_doc in enumerate(X):
            for tagged_sent in tagged_doc:
                for _, tag in tagged_sent:
                    if self.stack:
                        counter = self.pos_tag_counts_[y[idx]]
                    else:
                        counter = self.pos_tag_counts_["documents"]

                    if tag.startswith("N"):
                        counter["noun"] += 1
                    elif tag.startswith("J"):
                        counter["adjective"] += 1
                    elif tag.startswith("V"):
                        counter["verb"] += 1
                    # include particles with adverbs
                    elif tag.startswith("RB") or tag == "RP":
                        counter["adverb"] += 1
                    elif tag.startswith("PR"):
                        counter["pronoun"] += 1
                    elif tag.startswith("W"):
                        counter["wh- word"] += 1
                    elif tag == "CC":
                        counter["conjunction"] += 1
                    elif tag == "CD":
                        counter["digit"] += 1
                    # combine predeterminer and determiner
                    elif tag in ["DT" or "PDT"]:
                        counter["determiner"] += 1
                    elif tag == "EX":
                        counter["existential"] += 1
                    elif tag == "FW":
                        counter["non-English"] += 1
                    elif tag == "IN":
                        counter["preposition"] += 1
                    elif tag == "POS":
                        counter["possessive"] += 1
                    elif tag == "LS":
                        counter["list"] += 1
                    elif tag == "MD":
                        counter["modal"] += 1
                    elif tag in self.punct_tags:
                        counter["punctuation"] += 1
                    elif tag == "TO":
                        counter["infinitive"] += 1
                    elif tag == "UH":
                        counter["interjection"] += 1
                    elif tag == "SYM":
                        counter["symbol"] += 1
                    else:
                        counter["other"] += 1

    def draw(self, **kwargs):
        """
        Called from the fit method, this method creates the canvas and
        draws the part-of-speech tag mapping as a bar chart.

        Parameters
        ----------
        kwargs: dict
            generic keyword arguments.

        Returns
        -------
        ax : matplotlib axes
            Axes on which the PosTagVisualizer was drawn.
        """
        # Converts nested dict to nested list
        pos_tag_counts = np.array(
            [list(i.values()) for i in self.pos_tag_counts_.values()]
        )
        # stores sum of nested list column wise
        pos_tag_sum = np.sum(pos_tag_counts, axis=0)

        if self.frequency:
            # sorts the count and tags by sum for frequency true
            idx = (pos_tag_sum).argsort()[::-1]
            self._pos_tags = np.array(self._pos_tags)[idx]
            pos_tag_counts = pos_tag_counts[:, idx]

        if self.stack:
            bar_stack(
                pos_tag_counts,
                ax=self.ax,
                labels=list(self.labels_),
                ticks=self._pos_tags,
                colors=self.colors,
                colormap=self.colormap,
            )
        else:
            xidx = np.arange(len(self._pos_tags))
            colors = resolve_colors(
                n_colors=len(self._pos_tags), colormap=self.colormap, colors=self.colors
            )
            self.ax.bar(xidx, pos_tag_counts[0], color=colors)

        return self.ax

    def finalize(self, **kwargs):
        """
        Finalize the plot with ticks, labels, and title

        Parameters
        ----------
        kwargs: dict
            generic keyword arguments.
        """
        # NOTE: not deduping here, so this is total, not unique
        self.ax.set_ylabel("Count")

        if self.frequency:
            self.ax.set_xlabel(
                "{} part-of-speech tags, sorted by frequency".format(
                    self.tagset_names[self.tagset]
                )
            )
        else:
            self.ax.set_xlabel(
                "{} part-of-speech tags".format(self.tagset_names[self.tagset])
            )

        # bar stack(helper) sets the ticks if stack is true
        if not self.stack:
            self.ax.set_xticks(range(len(self._pos_tags)))
            self.ax.set_xticklabels(self._pos_tags, rotation=90)

        self.set_title(
            "PosTag plot for {}-token corpus".format(
                (sum([sum(i.values()) for i in self.pos_tag_counts_.values()]))
            )
        )

        # Call tight layout to maximize readability
        self.fig.tight_layout()

    def show(self, outpath=None, **kwargs):
        if outpath is not None:
            kwargs["bbox_inches"] = kwargs.get("bbox_inches", "tight")
        return super(PosTagVisualizer, self).show(outpath, **kwargs)


##########################################################################
## Quick Method
##########################################################################


def postag(
    X,
    y=None,
    ax=None,
    tagset="penn_treebank",
    colormap=None,
    colors=None,
    frequency=False,
    stack=False,
    parser=None,
    show=True,
    **kwargs,
):
    """
    Display a barchart with the counts of different parts of speech
    in X, which consists of a part-of-speech-tagged corpus, which the
    visualizer expects to be a list of lists of lists of (token, tag)
    tuples.

    Parameters
    ----------
    X : list or generator
        Should be provided as a list of documents or a generator
        that yields a list of documents that contain a list of
        sentences that contain (token, tag) tuples.

    y : ndarray or Series of length n
        An optional array of target values that are ignored by the
        visualizer.

    ax : matplotlib axes
        The axes to plot the figure on.

    tagset: string
        The tagset that was used to perform part-of-speech tagging.
        Either "penn_treebank" or "universal", defaults to "penn_treebank".
        Use "universal" if corpus has been tagged using SpaCy.

    colors : list or tuple of colors
        Specify the colors for each individual part-of-speech.

    colormap : string or matplotlib cmap
        Specify a colormap to color the parts-of-speech.

    frequency: bool {True, False}, default: False
        If set to True, part-of-speech tags will be plotted according to frequency,
        from most to least frequent.

    stack : bool {True, False}, default : False
        Plot the PosTag frequency chart as a per-class stacked bar chart.
        Note that fit() requires y for this visualization.

    parser : string or None, default: None
        If set to a string, string must be in the form of 'parser_tagger' or 'parser'
        to use defaults (for spacy this is 'en_core_web_sm', for nltk this is 'word').
        The 'parser' argument is one of the accepted parsing libraries. Currently
        'nltk' and 'spacy' are the only accepted libraries. NLTK or SpaCy must be
        installed into your environment. 'tagger' is the tagset to use. For example
        'nltk_wordpunct' would use the NLTK library with 'wordpunct' tagset. Or
        'spacy_en_core_web_sm' would use SpaCy with the 'en_core_web_sm' tagset.

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``

    kwargs : dict
        Pass any additional keyword arguments to the PosTagVisualizer.

    Returns
    -------
    visualizer: PosTagVisualizer
        Returns the fitted, finalized visualizer
    """
    # Instantiate the visualizer
    visualizer = PosTagVisualizer(
        ax=ax,
        tagset=tagset,
        colors=colors,
        colormap=colormap,
        frequency=frequency,
        stack=stack,
        parser=parser,
        **kwargs,
    )

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(X, y=y, **kwargs)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer object
    return visualizer



# yellowbrick.contrib
#
# Contains a variety of extra tools and experimental visualizers outside of
# core support or still in development.
#
#
# ID: __init__.py [a60bc41] nathan.danielsen@gmail.com $
